#include <windows.h>
#include <gl/gl.h>
#include <malloc.h>
#include <stdio.h>

#define	XAXISPEN	0	// used for drawing the X axis
#define YAXISPEN	1	// used for drawing the Y axis
#define ZAXISPEN	2	// used for drawing the Z axis
#define LABELPEN	3	// used for drawing the axes of non-highlighted parts

#define MARKPEN		4	// used for marking lines stuff
#define HIGHPEN		5	// used for highlighting lines
#define SHAFTPEN	6	// used for drawing the shaft and rise of parts
#define KNOBPEN		7	// used for drawing the knobs of parts

#define VERTPEN		8	// used for drawing the vertices
#define FINEPEN		9	// for the grid in parallel views
#define GRIDPEN		10	// for the coarse grid
#define EDGEPEN		11	// for interior edges

#define ORPHANPEN	12	// for edges with no bordering face
#define BORDERPEN	13	// for edges with one bordering face
#define SHARPPEN	14	// for marking sharp edges
#define NEWPEN		15	// for marking new stuff being added
#define PENCOUNT	16

#define X 0
#define Y 1
#define Z 2
#define LX 9
#define LY 10
#define LZ 11

#include "lsmtypes.h"
#include "calc.h"	// defines mathematical objects
#include "lists.h"	// defines linked list objects
#include "surf.h"	// defines texturing objects
#include "skel.h"	// defines articulation
#include "mesh.h"	// defines surface geometry

/*
typedef struct Normal {
	link	Link;
	TYPE	Type;
	INDEX	Index;
	short	Count;
	vert*	*Vertices;
//	vec	Spot;
} NORM, *PNORM;

#define NT_SMOOTH	0
#define NT_EDGE		1

// used for cross-referencing
typedef struct CrossRef {
	link	Link;
	FLAGS	Flags;
	INDEX	Index;
	PNORM	NormA,NormB,NormC;
} XREF, *PXREF;
*/

#include "view.h"

typedef struct Delta {
	link	Link;
	INDEX	Index;
	TYPE	DeltaType;
	TYPE	ObjectType;
	short	Len;	// number of words
	LSMWORD	*Data;	// the rest of the delta data
} DELTA, *PDELT;

#define	DT_ATOM	0	// break point between changes
#define	DT_ADD	1	// add selected data
#define	DT_CHG	2	// Replace
#define	DT_DEL	3	// remove indicated item

#define LSM0BLOCK	0x304d534c
#define LSM1BLOCK	0x314d534c
#define LSM2BLOCK	0x324d534c
#define	LSM3BLOCK	0x334d534c
#define OLDKNOBBLOCK	0x424f4e4b
#define OLDPARTBLOCK	0x54524150
#define OLDVERTBLOCK	0x54524556
#define OLDEDGEBLOCK	0x45474445
#define OLDFACEBLOCK	0x45434146
#define OLDQUADBLOCK	0x44415551
#define OLDFLAG0	0x4254412d
#define ATBASEFLAG	0x4254412b
#define OLDFLAG1	0x5048532d
#define SHARPFLAG	0x5048532b
#define LIMXBLOCK	0x584d494c
#define LIMYBLOCK	0x594d494c
#define LIMZBLOCK	0x5a4d494c
#define LIMLBLOCK	0x4c4d494c
#define NAMEBLOCK	0x4D4F4E24
#define FILEBLOCK	0x4C494624
#define OLDTINTBLOCK	0x544E4954
#define TINTATTR	0x544e542b
#define OLDUVPTBLOCK	0x54505655
#define UVMAPDATA	0x4D56552b
#define LOCKXBLOCK	0x584b4c2b
#define LOCKYBLOCK	0x594b4c2b
#define LOCKZBLOCK	0x5a4b4c2b
#define LOCKLBLOCK	0x4c4b4c2b
#define	HINGEFLAG	0x474e482b
#define	BALLFLAG	0x4c4c422b
#define LOCKVFLAG	0x564b4c2b

#define	EDGEDELT	0x4445
#define	FACEDELT	0x3346
#define	KNOBDELT	0x4e4b
#define PARTDELT	0x4150
#define QUADDELT	0x3446
#define	UVPTDELT	0x5055
#define TINTDELT	0x4954
#define	VERTDELT	0x4556

typedef struct Project {
	link	Link;	// we link these
	FLAGS	Flags;	// we have flags
	char	*SaveTitle;
	char	*SavePath;
	char	*ExportTitle;
	char	*ExportPath;
	view	cView;	// in use now
	view	mView;	// for use during the mesh view
	view	sView;	// for saving during zoom, drag, and rotate
	view	uView;	// for use during uv-map editing
	TYPE	ProjectType;	// the project type
	short	EditMode;		// the editing mode in force in the project
	short	MouseMode;		// the mouse mode in force for the project
	vec	UserScale,UserRotate,UserTranslate;
	matrix	UserMatrix;
	chain	Deltas;
	PDELT	CurrentDelta;
} PROJECT,*PPROJECT;

#define PF_MODIFIED		2	// modified since last load
#define PF_SHOWAXES		4	// axes will be visible
#define PF_INVALIDVIEW	16
#define PF_NOTDEFAULTNAME	32

#define MF_MESHCHANGED	256
#define MF_SHOWPARTS	512
#define MF_SHOWVERTS	1024
#define MF_SHOWEDGES	2048
#define MF_SHOWFACES	4096

#define	EM_MESHEDIT	0
#define EM_MAPEDIT	256
#define EM_SKELETON	512
#define PT_MODEL		0	//

typedef union {
	edge* e;
	face* f;
	knob* k;
	part* p;
	tint* t;
	vert* v;
	uvpt* u;
} MYPTR;

typedef struct Model {
	PROJECT	Project;
	chain	Knobs,Parts,Vertices,Edges,Faces,Tints,UVPts /* ,Normals,CrossRefs */ ;
	INDEX	NextEdgeIndex,NextFaceIndex,NextKnobIndex,NextPartIndex,
 				NextTintIndex,NextUVPtIndex,NextVertIndex;
	tint	*defTint,*CurTint;
	float	MeshSnap,MapSnap;	// the current grid snap precision
} MODEL, *PMODEL;

#define FIRSTVERT(m) ((vert*)(((PMODEL)m)->Vertices.Next))
#define STEPVERTS(m,v) for(v=FIRSTVERT(m);NEXT(v)!=NULL;v=(vert*)NEXT(v))

#define FIRSTPART(m) ((part*)(((PMODEL)m)->Parts.Next))
#define STEPPARTS(m,p) for(p=FIRSTPART(m);NEXT(p)!=NULL;p=(part*)NEXT(p))

#define FIRSTKNOB(m) ((knob*)(((PMODEL)m)->Knobs.Next))
#define STEPKNOBS(m,k) for(k=FIRSTKNOB(m);NEXT(k)!=NULL;k=(knob*)NEXT(k))

#define FIRSTEDGE(m) ((edge*)(((PMODEL)m)->Edges.Next))
#define STEPEDGES(m,e) for(e=FIRSTEDGE(m);NEXT(e)!=NULL;e=(edge*)NEXT(e))

#define FIRSTFACE(m) ((face*)(((PMODEL)m)->Faces.Next))
#define STEPFACES(m,f) for(f=FIRSTFACE(m);NEXT(f)!=NULL;f=(face*)NEXT(f))

#define FIRSTTINT(m) ((tint*)(((PMODEL)m)->Tints.Next))
#define STEPTINTS(m,t) for(t=FIRSTTINT(m);NEXT(t)!=NULL;t=(tint*)NEXT(t))

#define FIRSTUVPT(m) ((uvpt*)(((PMODEL)m)->UVPts.Next))
#define STEPUVPTS(m,u) for(u=FIRSTUVPT(m);NEXT(u)!=NULL;u=(uvpt*)NEXT(u))

#define FIRSTGANG(m) ((gang*)(((PMODEL)m)->Gangs.Next))
#define STEPGANGS(m,g) for(g=FIRSTGANG(m);NEXT(g)!=NULL;g=(gang*)NEXT(g))

#define FIRSTDELT(p) ((PDELT)(((PPROJECT)p)->Deltas.Next))
#define STEPDELTS(p,d) for(d=FIRSTDELT(p);NEXT(d)!=NULL;d=(PDELT)NEXT(d))

//#define FIRSTNORM(m) ((PNORM)(((PMODEL)m)->Normals.Head))
//#define STEPNORMS(m,n) for(n=FIRSTNORM(m);NEXT(n)!=NULL;n=(PNORM)NEXT(n))

#define VCOUNT(m) (((PMODEL)m)->Vertices.Count)
#define FCOUNT(m) (((PMODEL)m)->Faces.Count)
#define ECOUNT(m) (((PMODEL)m)->Edges.Count)

// mouse modes
#define	MM_NOTHING	0	// user wasn't doing anything
#define	MM_RIGHTDOWN	1
#define	MM_RIGHTBOX 2
#define	MM_ADDVERT	3	// have just given add vertices command
#define	MM_DRAGVERT 5
#define	MM_ADDPART0 6	// have just given add parts command
#define	MM_ADDPART1 7
#define	MM_ADDPART2 8
#define	MM_ADDPARTX 9
#define	MM_DRAGKNOB 10
#define	MM_DRAGKNOBX	11
#define	MM_HOOKHIGH	12
#define	MM_LEFTDOWN	13
#define	MM_GRABVERT	14
#define	MM_HOOKVERT	15
#define	MM_HOOKVERX 16
#define	MM_DBLCLICK	17
#define	MM_ADDMESH0	18
#define	MM_ADDMESH1	19
#define	MM_ADDMESH2 20
#define	MM_ADDMESHX 21
#define	MM_GRABVERX 22
#define	MM_DBLCLICX 23
#define MM_LEFTBOX 24
#define MM_GRABuvpt	25
#define MM_DRAGuvpt	26
#define MM_DRAGRISE	27
#define MM_DRAGRISEX	28
#define MM_PANVIEW0	29
#define MM_PANVIEW1	30
#define MM_ZOOMVIEW0	31
#define MM_ZOOMVIEW1	32
#define MM_ROTVIEW0	33
#define MM_ROTVIEW1	34

// edit actions (implements multiple functionality of transform functions)
#define	EA_NADA	0
#define	EA_EXTRUDE	1
#define	EA_COPY	2

#define HT_NOTHING	0
#define HT_CLEAR	1	// do not preserve highlighting from last check
#define HT_FACE	2
#define HT_EDGE	4
#define HT_VERT	8
#define	HT_PART	16
#define HT_KNOB	32
#define HT_UVPT	64
#define	HT_GANG	128
#define HT_NORISE	256

#define HALFPI		 1.570796326795
#define PI			 3.14159265359
#define LN10		 2.302585092994
#define RAD2DEG		57.29577951308
#define DEG2RAD		 0.01745329251994
#define SQRT_3_4THS	 0.86602540378444

// for setting up dialog boxes
typedef struct DialogData {
	short type;	// the type
	int	item;	// the dialog item (or count of items)
	union {		// the data in question
		char	*tdata;
		int		idata;
		float	fdata;
		float	*vdata;
		COLORREF	cdata;
		MYPTR	mdata;
	} data;
}	DLGX,*PDLGX;

#define	DXT_INT	0	// an integer
#define	DXT_TXT	1	// a string
#define	DXT_NUM	2	// a float
#define	DXT_RGB	3	// a COLORREF
#define	DXT_TTL	4	// the title for the dialog box
#define	DXT_VEC	5	// the pointer to an XYZ vector
#define	DXT_FLG	6	// a flag
#define DXT_OBJ	7	// the object being edited

// and now to declare some functions

// delta.c
void AddAtomDelta(PPROJECT);
PDELT CreateDelta(PPROJECT,TYPE,TYPE,int);
void DestroyDelta(PDELT);
void ExtendDelta(PDELT,int,LSMWORD *);
PDELT NextAtomDelta(PPROJECT);
PDELT NextMatch(PDELT);
PDELT PreviousAtomDelta(PPROJECT);
PDELT PreviousMatch(PDELT);
void RebuildObjectByDelta(PPROJECT,PDELT);
void RemoveObjectByDelta(PPROJECT,PDELT);
void ReplaceObjectByDelta(PPROJECT,PDELT);
void ClipDeltaChain(PPROJECT);

//dialogs.c
BOOL _stdcall AboutDlg(HWND,UINT,WPARAM,LPARAM);
BOOL _stdcall EditPartDlg(HWND,UINT,WPARAM,LPARAM);
BOOL _stdcall InputMatrixDlg(HWND,UINT,WPARAM,LPARAM);
BOOL _stdcall EditTintDlg(HWND,UINT,WPARAM,LPARAM);
BOOL _stdcall PickTintDlg(HWND,UINT,WPARAM,LPARAM);
BOOL _stdcall InputStringDlg(HWND,UINT,WPARAM,LPARAM);
void ReportSystemError(char *);
void InitializeDialogData(HWND,PDLGX);
void ReturnDialogData(HWND,PDLGX);
void DestroyDialogData(PDLGX);
edge*GetDialogObject(PDLGX);
void SelectDialogTint(HWND,tint*);

//display.c
void DrawAxes(void);
void DrawEdges(void);
void DrawFaces(void);
void DrawHighlight(void);
void DrawKnobs(void);
void DrawLSMEdge(edge*);
void DrawMappedPolygon(float*,float*,float*,float*,float*,float*,float*,float*);
void DrawMarking(void);
void DrawParts(void);
void DrawUVEdges(void);
void DrawUVPt(uvpt*);
void DrawUVPts(void);
void DrawVertices(void);
void InvalidateView(void);
void Jump2Origin(void);
void PaintFace(face*);
void PaintMapBackground(void);
void RedoDisplay(void);
void RefreshUVPt(uvpt*);
void ResizeDisplay(void);
void RightRect(RECT *);
void SelectGLColor(COLORREF);
void ValidateMesh(PMODEL);
void ViewXY(void);
void ViewXZ(void);
void ViewYZ(void);
void Zoom2Fit(PPROJECT);
void ZoomInTenPercent(void);
void ZoomOutTenPercent(void);

//edge.c
edge* CreateEdge(PMODEL,vert*,vert*,BOOL);
PDELT CreateEdgeDelta(edge*,TYPE);
void DestroyEdge(edge*,BOOL);
edge* FindEdge(PMODEL,vert*,vert*);
void FinishEdgeByDelta(PMODEL,edge*,PDELT);
edge* GetEdge(PMODEL,vert*,vert*,BOOL);
edge* FindDuplicate(PMODEL,edge*);

//edit.c
void AttachMarked(part*);
void AttachVertex(vert*,part*);
void BlankSelection(void);
void CheckEdgeMark(vert*);
void ClearAll(void);
void CopyMarked();
void DecSelection(void);
void DeleteEdges(BOOL);
void DeleteFaces(BOOL);
void DeleteHighlight(BOOL);
void DeleteVerts(BOOL);
void DetachHighlight(BOOL);
void DetachMarked(void);
void EditHighlight(HWND);
void ExtrudeMarked(void);
void IncSelection(void);
void MarkAll(void);
void MatrixMarked(HWND,short);
//void MergeMarked(void);
void RotateMarked(HWND,short);
void ScaleMarked(HWND,short);
void SetGridSnap(HWND);
void SharpenEdges(void);
void SharpenVertices(void);
void SmoothEdges(void);
void SmoothVertices(void);
void ToggleAll(void);
void ToggleHighlight(void);
void TranslateMarked(HWND,short);
void AlignMarked(HWND,int);
void LockHighlight(void);
void LockMarked(void);

//face.c
face* CreateFace(PMODEL,vert*,vert*,vert*,vert*,BOOL);
PDELT CreateFaceDelta(face*,TYPE);
void DestroyFace(face*,BOOL);
void FinishFaceByDelta(PMODEL,face*,PDELT);
tint* GetFaceTint(face*);
void OrderAllVertices(PMODEL);
void OrderFaceVertices(face*);
face* OtherSide(edge*,face*);

//files.c
void ExportProject(void);
void GetBMPFileName(HWND,int);
void ImportProject(void);
void OpenFromCommandLine(LPSTR);
void OpenProject(void);
float ReadFloat(HANDLE);
long ReadLong(HANDLE);
void ReadPaddedString(HANDLE,char *,int);
short ReadShort(HANDLE);
void SaveProject(BOOL);
void WriteFloat(HANDLE,float);
void WriteFloatString(HANDLE,float);
void WriteHexString(HANDLE,int);
void WriteIntString(HANDLE,int);
void WriteLong(HANDLE,long);
void WritePaddedString(HANDLE,char *);
void WriteShort(HANDLE,short);
void WriteString(HANDLE,char*);
void WriteStrings(HANDLE,...);
void WriteUVecString(HANDLE,float*);
void WriteVecString(HANDLE,float*);

//inc_file.c
void WriteINCFile(HANDLE,PMODEL);
void WritePartBuild(HANDLE,part*);

//knob.c
knob* CopyKnob(knob*,BOOL);
knob* CreateKnob(PMODEL,SHORT,SHORT,BOOL);
PDELT CreateKnobDelta(knob*,TYPE);
void DestroyKnob(knob*,BOOL);
void FinishKnobByDelta(PMODEL,knob*,PDELT);
void MoveKnobEnd(knob*);
void ReplaceKnob(knob*,knob*);

// lsm_file.c
void WriteLSMFile(HANDLE,PMODEL);
PMODEL ReadLSMFile(HANDLE);
void ReadLSM0File(HANDLE,PMODEL);
void ReadLSM1File(HANDLE,PMODEL);
void WriteKnobBlock(HANDLE,knob*);
void WritePartBlock(HANDLE,part*);
void WriteVertBlock(HANDLE,vert*);
void WriteEdgeBlock(HANDLE,edge*);
void WriteFaceBlock(HANDLE,face*);
void WriteTintBlock(HANDLE,tint*);
void WriteUVPtBlock(HANDLE,uvpt*);
long ReadKnobBlock(HANDLE,PMODEL);
long ReadPartBlock(HANDLE,PMODEL);
long ReadVertBlock(HANDLE,PMODEL);
long ReadEdgeBlock(HANDLE,PMODEL);
long ReadFaceBlock(HANDLE,PMODEL);
long ReadQuadBlock(HANDLE,PMODEL);
long ReadPaintInfo(HANDLE,PMODEL,face*);
long ReadTintBlock(HANDLE,PMODEL);
long ReadUVPtBlock(HANDLE,PMODEL);
long SkipBlock(HANDLE);
void WriteDelta(HANDLE,PDELT);

// main.c
void UpdateStatusBar(LPSTR,WORD,WORD);

// math.c
void MatrixMultiply(float*,float*,float*);
void MatrixInvert(float*,float*);
//void RedoMapMatrices(int,int);
float Normalize(float*,float*);
void SubtractVector(float*,float*,float*);
void TransformVector(float*,float*,float*);
void InvertVector(float*,float*,float*);
void InvertUVector(float*,float*,float*);
float Distance2(float*,float*);
float Distance(float*,float*);
void Polar2Cart(part*,float*);
void Cart2Polar(part*,float*);
void CalcModelAngles(PMODEL);
BOOL isPowerof2(UINT);
UINT nearestPowerof2(UINT);
void MapXYZ2UV(float*,float*);

//menu.c
HMENU BuildTintPopup(PMODEL);
void UpdateModelMenu(void);

//model.c
PMODEL CreateModel(void /*char * */);
void DestroyModelInfo(PMODEL);
PMODEL FindModel(link *);
void ReindexModelObjects(PMODEL);
edge* IndexEdge(PMODEL,INDEX);
face* IndexFace(PMODEL,INDEX);
knob* IndexKnob(PMODEL,INDEX);
part* IndexPart(PMODEL,INDEX);
tint* IndexTint(PMODEL,INDEX);
uvpt* IndexUVPt(PMODEL,INDEX);
vert* IndexVert(PMODEL,INDEX);

// mouse.c
void AttachHighlight(BOOL);
void CheckHighlight(short);
void ClearHighlight(void);
void LeftClick(HWND,WPARAM,LPARAM);
void LeftDoubleClick(HWND,WPARAM,LPARAM);
void LeftRelease(HWND,WPARAM,LPARAM);
void MouseMove(HWND,WPARAM,LPARAM);
void RightClick(HWND,WPARAM,LPARAM);
void RightRelease(HWND,WPARAM,LPARAM);
void PrepSingleKnobDrag(knob*);
void PrepVertDrag(void);
void PrepSingleVertDrag(vert*);
void FinishVertDrag(void);
void FinishKnobDrag(void);
void SetMouseMode(short);
void CheckSelectBox(RECT *,short);

// normal.c
//PNORM CreateNormal(PMODEL);
//void DestroyNormal(PNORM,BOOL);
//PXREF CreateCrossRef(PMODEL);
//void DestroyCrossRef(PXREF,BOOL);
//PNORM FindNormal(face*,vert*);
//PNORM FindNormalBoth(face*,vert*);
//PNORM FindNormalBack(face*,vert*,face*);
//PNORM FindNormalFore(face*,vert*,face*);
//PNORM BuildNormal(face*,vert*);
//edge* StepEdgeBack(face*,vert*);
//edge* StepEdgeFore(face*,vert*);
//face* OtherSide(face*,edge*);
//vert* OtherEnd(edge*,vert*);
//void BuildAllNormals(PMODEL);
//void CalcNormal(PNORM,float *);
//void DestroyAllNormals(PMODEL);

// opengl.c
HGLRC PrepareRenderingContext(HDC);
void OpenGLError(void);

//part.c
void BaseParts(PMODEL);
void CalcRise(part*);
void Constrain(knob*);
void ConvertJoint(part*,TYPE);
part* CreatePart(PMODEL,knob*,knob*,knob*,BOOL);
PDELT CreatePartDelta(part*,TYPE);
void DestroyPart(part*,BOOL);
void FinishPartByDelta(PMODEL,part*,PDELT);
BOOL hasNonUniChildren(PMODEL,part*);
void RebuildPartMatrices(void);
void RebuildAllMatrices(PMODEL);
void UnleashPart(part*,BOOL);

//project.c
PPROJECT CreateProject(void);
BOOL DestroyProject(PPROJECT);
void ActivateFirstProject(void);
void ActivateNextProject(void);
void ActivatePrevProject(void);
void ActivateProject(PPROJECT);
void DestroyAllDeltas(PPROJECT);
BOOL DestroyAllProjects(void);
void InitProject(PPROJECT);
void CloseActiveProject(void);
void EnterMode(short);
void SaveView(void);
PDELT PreviousAtomDelta(PPROJECT);
PDELT NextAtomDelta(PPROJECT);

//scroll.c
void SetScroll(void);
void HScroll(WPARAM);
void VScroll(WPARAM);

//string.c
void FatFarm(char *);
void CopyString(char *,char *);
void DestroyString(char *);
char *CloneString(char *);
void SetWindowTitle(void);
char *CreateModelName(void);
char *ConcatStrings(char*,char *);
char *CreatePartName(PMODEL);
char *CreateTintName(PMODEL);
BOOL IsUniqueProjectName(char *);
BOOL IsUniquePartName(PMODEL,char *,part*);
BOOL IsUniqueTintName(PMODEL,char *,tint*);
void ScrubString(char *,char,char,char *);

//tint.c
tint* CreateTint(PMODEL,COLORREF,BOOL);
PDELT CreateTintDelta(tint*,TYPE);
void DestroyTint(tint*,BOOL);
void FinishTintByDelta(PMODEL,tint*,PDELT);
void EditFaceTint(void);
void BuildTintBitmap(HDC,tint*);
void IndexTints(PMODEL);
void TintWindow(HWND,tint*);
void CurTintHighFace(void);
void ApplyFaceTint(face*,tint*);
void GetTintBitmap(tint*);
void FillWindow(HWND,COLORREF);
void DeriveSolid(tint*);

//toolbar.c
HWND CreateAToolBar(HWND,HINSTANCE);
void UpdateToolBar(void);

// undo.c
void Redo(PPROJECT);
void Undo(PPROJECT);

//uvpt.cpp
void AddMapping(PMODEL,face*,BOOL);
void CompleteUVData(face*);
uvpt* CopyUVPt(uvpt*);
uvpt* CreateUVPt(PMODEL,BOOL);
PDELT CreateUVPtDelta(uvpt*,TYPE);
void DestroyUVPt(uvpt*,BOOL);
void ExtendMapping(face*,edge*,float*);
void FinishUVPtByDelta(PMODEL,uvpt*,PDELT);
void PullUVData(face*,edge*);
void RemMapping(face*,BOOL);
void RemoveUnusedUVPts(PMODEL);

//vertex.c
vert* CreateVertex(PMODEL,short,short,BOOL);
PDELT CreateVertDelta(vert*,TYPE);
void DestroyVertex(vert*,BOOL);
void FinishVertByDelta(PMODEL,vert*,PDELT);
void ReplaceVertex(vert*,vert*);
vert* SelectVert(PMODEL,INDEX);
void SnapVertex(vert*);
void DetachVertex(vert*,BOOL);

