#include "data.h"

extern PPROJECT ActiveProject;

void AddAtomDelta(PPROJECT p) {
	if(p->CurrentDelta)
	if(p->CurrentDelta->DeltaType==DT_ATOM) return;
	CreateDelta(p,DT_ATOM,0,0);
}

PDELT CreateDelta(PPROJECT p,TYPE d_type,TYPE o_type,int words) {
	PDELT d;

	if(p==NULL) p=ActiveProject;
	if(p==NULL) return NULL;

	ClipDeltaChain(p);

	d=(PDELT)malloc(sizeof(DELTA));
	if(d) {
		addTail(&p->Deltas,(link*)d);
		d->DeltaType=d_type;
		d->ObjectType=o_type;
		d->Len=words;
		d->Index=0;
		if(words==0 || d_type==DT_DEL || d_type==DT_ATOM)	{
			d->Data=(LSMWORD*)0;
		} else {
			d->Data=(LSMWORD*)malloc(words*sizeof(LSMWORD));
			if(d->Data==NULL) {
				ReportSystemError("Data malloc for delta object failed.");
				remLink(&d->Link);
				free(d);
				d=NULL;
			}
		}
		p->CurrentDelta=d;
	}

	return d;
}

void DestroyDelta(PDELT d) {
	if(d==NULL) {
		ReportSystemError("NULL passed to DestroyDelta()");
		return;
	}
	remLink(&d->Link);
	if(d->Data) free(d->Data);
	free(d);
}

void ExtendDelta(PDELT d,int wc,LSMWORD *add) {
	LSMWORD *total;
	int i;

	total=(LSMWORD *)malloc((wc+d->Len)*sizeof(LSMWORD));
	if(total==NULL) {
		ReportSystemError("Extension of Delta object failed.");
		return;
	}

	if(d->Data) {
		for(i=0;i<d->Len;i++) total[i]=d->Data[i];
		free(d->Data);
	}

	for(i=0;i<wc;i++) total[d->Len+i]=add[i];
	d->Len+=wc;

	d->Data=total;
}

void RebuildObjectByDelta(PPROJECT p,PDELT d) {
	MYPTR mp;

	switch(d->ObjectType) {
		case UVPTDELT:
			mp.u=CreateUVPt((PMODEL)p,FALSE);
			FinishUVPtByDelta((PMODEL)p,mp.u,d);
		break;
		case VERTDELT:
			mp.v=CreateVertex((PMODEL)p,0,0,FALSE);
			FinishVertByDelta((PMODEL)p,mp.v,d);
		break;
		case KNOBDELT:
//			ReportSystemError("Rebuilding knob");
			mp.k=CreateKnob((PMODEL)p,0,0,FALSE);
			FinishKnobByDelta((PMODEL)p,mp.k,d);
		break;
		case EDGEDELT:
			mp.e=CreateEdge((PMODEL)p,IndexVert((PMODEL)p,d->Data[0].l),
				IndexVert((PMODEL)p,d->Data[1].l),FALSE);
			FinishEdgeByDelta((PMODEL)p,mp.e,d);
		break;
		case PARTDELT:
			mp.p=CreatePart((PMODEL)p,IndexKnob((PMODEL)p,d->Data[0].l),
				IndexKnob((PMODEL)p,d->Data[1].l),
				IndexKnob((PMODEL)p,d->Data[2].l),FALSE);
			FinishPartByDelta((PMODEL)p,mp.p,d);
		break;

		case QUADDELT:
			mp.f=CreateFace((PMODEL)p,IndexVert((PMODEL)p,d->Data[0].l),
				IndexVert((PMODEL)p,d->Data[1].l),
				IndexVert((PMODEL)p,d->Data[2].l),
				IndexVert((PMODEL)p,d->Data[3].l),FALSE);
			FinishFaceByDelta((PMODEL)p,mp.f,d);
		break;
		case FACEDELT:
			mp.f=CreateFace((PMODEL)p,IndexVert((PMODEL)p,d->Data[0].l),
				IndexVert((PMODEL)p,d->Data[1].l),
				IndexVert((PMODEL)p,d->Data[2].l),NULL,FALSE);
			FinishFaceByDelta((PMODEL)p,mp.f,d);
		break;

		case TINTDELT:
			mp.t=CreateTint((PMODEL)p,0,FALSE);
			FinishTintByDelta((PMODEL)p,mp.t,d);
		break;
	}
}

void RemoveObjectByDelta(PPROJECT p,PDELT d) {
//	MYPTR mp;
	switch(d->ObjectType) {
		case VERTDELT:	DestroyVertex(IndexVert((PMODEL)p,d->Index),FALSE);	break;
		case EDGEDELT:	DestroyEdge(IndexEdge((PMODEL)p,d->Index),FALSE);	break;
		case QUADDELT:
		case FACEDELT:	DestroyFace(IndexFace((PMODEL)p,d->Index),FALSE);	break;
		case PARTDELT:	DestroyPart(IndexPart((PMODEL)p,d->Index),FALSE);	break;
		case KNOBDELT:	DestroyKnob(IndexKnob((PMODEL)p,d->Index),FALSE);	break;
		case UVPTDELT:	DestroyUVPt(IndexUVPt((PMODEL)p,d->Index),FALSE);	break;
		case TINTDELT:	DestroyTint(IndexTint((PMODEL)p,d->Index),FALSE);	break;
	}
}

void ReplaceObjectByDelta(PPROJECT p,PDELT d) {
	MYPTR mp;

	switch(d->Data[0].l) {
		case UVPTDELT:
			mp.u=IndexUVPt((PMODEL)p,d->Index);
			FinishUVPtByDelta((PMODEL)p,mp.u,d);
		break;
		case VERTDELT:
			mp.v=IndexVert((PMODEL)p,d->Index);
			FinishVertByDelta((PMODEL)p,mp.v,d);
		break;
		case KNOBDELT:
			mp.k=IndexKnob((PMODEL)p,d->Index);
			FinishKnobByDelta((PMODEL)p,mp.k,d);
		break;
		case EDGEDELT:
			mp.e=IndexEdge((PMODEL)p,d->Index);
			FinishEdgeByDelta((PMODEL)p,mp.e,d);
		break;
		case PARTDELT:
			mp.p=IndexPart((PMODEL)p,d->Index);
			FinishPartByDelta((PMODEL)p,mp.p,d);
		break;
		case QUADDELT:
		case FACEDELT:
			mp.f=IndexFace((PMODEL)p,d->Index);
			FinishFaceByDelta((PMODEL)p,mp.f,d);
		break;
		case TINTDELT:
			mp.t=IndexTint((PMODEL)p,d->Index);
			FinishTintByDelta((PMODEL)p,mp.t,d);
		break;
	}
}

void ClipDeltaChain(PPROJECT p) {
	PDELT d,nd;

	if(p->CurrentDelta==NULL) return;

	d=(PDELT)(p->CurrentDelta->Link.Next);

	while(d->Link.Next!=NULL) {
		nd=(PDELT)(d->Link.Next);
		DestroyDelta(d);
		d=nd;
	}
}

