#include "data.h"
#include "lsmres.h"

extern PPROJECT ActiveProject;
extern HWND hwndMain;

char FullPathBuffer[256];
char TitleBuffer[40];
char BMPFilter[]={ "Bitmap files (*.bmp)\0*.bmp\0" };
char ProjectFilter[]={ "LionSnake model (*.lsm)\0*.lsm\0" };
char ModelFilter[]={ "LionSnake model (*.lsm)\0*.lsm\0" };
char ModelExportFilter[]= {
    "POV-Ray posable include file (*.inc)\0*.inc\0"
};

char *ModelExt="lsm";
BOOL EndOfFile=FALSE;
char InputBuffer[64];

#define PI_MODEL 1
#define MEFI_INC 1

OPENFILENAME FileData= {
 sizeof(OPENFILENAME),	// lStructSize, never changes
 (HWND) NULL,		// hwndOwner, set to main window
 (HINSTANCE) NULL,	// hInstance, not used
 (LPCTSTR) NULL,		// lpstrFilter, indicates file type filter to use
 (LPTSTR) NULL,		// lpstrCustomFilter, not used
 (DWORD) 0,			// nMaxCustomFilter, not used
 (DWORD) 0,			// nFilterIndex, returns index of selected filter type
 (LPTSTR) FullPathBuffer,	// lpstrFile, full path of selected file
 (DWORD) 256,		// nMaxFile, size of buffer for full path
 (LPTSTR) TitleBuffer,	// lpstrFileTitle, buffer to receive file title
 (DWORD) 40,		// nMaxFileTitle, size of file title buffer
 (LPCTSTR) NULL,		// lpstrInitialDir, not used
 (LPCTSTR) NULL,		// lpstrTitle, for specifying dialog box window title
 (DWORD) NULL,		// Flags, set by code
 (WORD) 0,			// nFileOffset, offset of file name within returned string;
 (WORD) 0,			// nFileExtension, offset of extension within returned string;
 (LPCTSTR) NULL,		// lpstrDefExt, pointer to default extension;
 (LPARAM) 0,		// lCustData, probably not used;
 (LPOFNHOOKPROC) NULL,	// lpfnHook, probably not used;
 (LPCTSTR) NULL		// lpTemplateName, probably not used;
};

// following fields are used in getting file names:
// FileData.hwndOwner, set to main window
// FileData.lpstrFilter, indicates file type filter to use
// FileData.nFilterIndex, returns index of selected filter type
// FileData.lpstrTitle, for specifying dialog box window title
// FileData.Flags, set by code
// FileData.nFileOffset, offset of file name within returned string;
// FileData.nFileExtension, offset of extension within returned string;
// FileData.lpstrDefExt, pointer to default extension;

// Flags OFN_CREATEPROMPT ask to create if file does not exist
// OFN_DONTADDTORECENT prevents addition to recent documents list
// OFN_FILEMUSTEXIST prevents specification of non-existent file
// OFN_HIDEREADONLY hides the read-only check box.
// OFN_OVERWRITEPROMPT prompts user to approve overwriting if file exists
// OFN_EXTENSIONDIFFERENT user specified a different extension other than default

// <header> { <data> }

void SaveProject(BOOL getname) {
	HANDLE fhand;
// getname is TRUE when user selects "Save As"
// PF_NOTDEFAULTNAME is set when the user has saved the file at least once
	if(getname ||(ActiveProject->Flags&PF_NOTDEFAULTNAME)==0) {
		FileData.hwndOwner=hwndMain;
// should set the filter according to the project type
		FileData.lpstrFilter=ModelFilter;
		FileData.nFilterIndex=1;
		FileData.lpstrTitle=NULL;
		FileData.Flags=OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
// this should also be set according to the project type
		FileData.lpstrDefExt=ModelExt;
		strcpy(FullPathBuffer,ActiveProject->SaveTitle);
		if(GetSaveFileNameA(&FileData)==0) return;

		DestroyString(ActiveProject->SavePath);
		ActiveProject->SavePath=CloneString(FullPathBuffer);
		ActiveProject->Flags|=PF_NOTDEFAULTNAME;
	} else {
		strcpy(FullPathBuffer,ActiveProject->SavePath);
	}

	fhand=(HANDLE)CreateFile(FullPathBuffer,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	if(fhand==INVALID_HANDLE_VALUE) {
		ReportSystemError("Couldn't create the file");
		return;
	}

	switch(ActiveProject->ProjectType) {
	case PT_MODEL:
		DestroyAllDeltas(ActiveProject);
		ReindexModelObjects((PMODEL)ActiveProject);
		CalcModelAngles((PMODEL)ActiveProject);
		WriteLSMFile(fhand,(PMODEL)ActiveProject);
		AddAtomDelta(ActiveProject);
	break;
	}

	CloseHandle(fhand);

	ActiveProject->Flags|=PF_NOTDEFAULTNAME;

// name project to what the user specified
	DestroyString(ActiveProject->SaveTitle);
	ActiveProject->SaveTitle=CloneString(FullPathBuffer+FileData.nFileOffset);
	ActivateProject(ActiveProject);
}

void ExportProject(void) {
	 HANDLE fhand;

	FileData.hwndOwner=hwndMain;
// filter should be set according to the project type
	FileData.lpstrFilter=ModelExportFilter;
	FileData.nFilterIndex=1;
	FileData.lpstrTitle="Export";
	FileData.Flags=OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
	FileData.lpstrDefExt=NULL; // no default for exporting
	FullPathBuffer[0]=0;

	if(GetSaveFileNameA(&FileData)==0) return;

	if(FileData.nFileExtension==0) {
//		ReportSystemError("Ending file name with period.");
		FileData.nFileExtension=strlen(FullPathBuffer);
		if(FullPathBuffer[FileData.nFileExtension-1]=='.') {
			FileData.nFileExtension--;
			FullPathBuffer[FileData.nFileExtension]='\0';
		}
	}

	if(FullPathBuffer[FileData.nFileExtension]=='\0') {
	switch(FileData.nFilterIndex) {
		case MEFI_INC:
			FullPathBuffer[FileData.nFileExtension]='.';
			FullPathBuffer[FileData.nFileExtension+1]='i';
			FullPathBuffer[FileData.nFileExtension+2]='n';
			FullPathBuffer[FileData.nFileExtension+3]='c';
			FullPathBuffer[FileData.nFileExtension+4]='\0';
		break;
	}
	}

	fhand=(HANDLE)CreateFile(FullPathBuffer,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	if(fhand==INVALID_HANDLE_VALUE) {
		ReportSystemError("Couldn't create the file");
		return;
	}

	switch(FileData.nFilterIndex) {
//	case MEFI_MSH:
//		WriteMSHFile(fhand,(PMODEL)ActiveProject);
//	break;
//	case MEFI_PCM:
//		WritePCMFile(fhand,(PMODEL)ActiveProject);
//	break;
	case MEFI_INC:
		WriteINCFile(fhand,(PMODEL)ActiveProject);
	break;
	}

	CloseHandle(fhand);
}

void OpenProject(void) {
	HANDLE fhand;
	PMODEL m;

// following fields are used in getting file names:
	FileData.hwndOwner=hwndMain;
	FileData.lpstrFilter=ProjectFilter;
	FileData.nFilterIndex=1;
	FileData.lpstrTitle=NULL;
	FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

	if(GetOpenFileNameA(&FileData)==0) return;

	fhand=(HANDLE)CreateFile(FullPathBuffer,GENERIC_READ,0,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	if(fhand==INVALID_HANDLE_VALUE) {
		ReportSystemError("Could not open file");
		return;
	}
	EndOfFile=FALSE;

	// determine file type
	switch(FileData.nFilterIndex) {
	case PI_MODEL:
		m=ReadLSMFile(fhand);
		if(m!=NULL) {
			m->Project.Flags|=PF_NOTDEFAULTNAME;
			ActivateProject((PPROJECT)m);
			ValidateMesh(m);
			SendMessage(hwndMain,WM_PAINT,0,0);
			Zoom2Fit((PPROJECT)m);
		}
	break;
	}
	CloseHandle(fhand);
}

void OpenFromCommandLine(LPSTR fname) {
	HANDLE fhand;
	PMODEL m;
	int length;
	char *nname,*qname;

	length=strlen(fname);
	if(length==0) {
		CreateModel();
		ActivateFirstProject();
		return;
	}

	nname=(char *)malloc(length+1);
	ScrubString((char *)fname,'"','\0',nname);
//	strrepl(fname,"\"",NULL,nname);
	GetFullPathNameA(nname,256,FullPathBuffer,&qname);
	strcpy(TitleBuffer,qname);

	fhand=(HANDLE)CreateFile(FullPathBuffer,GENERIC_READ,0,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
	free(nname);
	if(fhand==INVALID_HANDLE_VALUE) {
		ReportSystemError("Could not open file");
		return;
	}

	if(strstr(fname,".lsm")!=0) {
		EndOfFile=FALSE;
		m=ReadLSMFile(fhand);
		if(m!=NULL) {
			m->Project.Flags|=PF_NOTDEFAULTNAME;
			ActivateProject((PPROJECT)m);
			ValidateMesh(m);
			SendMessage(hwndMain,WM_PAINT,0,0);
			Zoom2Fit((PPROJECT)m);
		} else {
			ReportSystemError("The specified file was invalid.");
		}
		CloseHandle(fhand);
		return;
	}
}

/*
void ImportProject(void) {
	fstream& fhand;
	PPROJECT p;

// following fields are used in getting file names:
	FileData.hwndOwner=hwndMain;
	FileData.lpstrFilter=ImportFilter;
	FileData.nFilterIndex=1;
	FileData.lpstrTitle="Import";
	FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
	FileData.lpstrDefExt=NULL; // no default for importing
	FullPathBuffer[0]=0;

	if(GetOpenFileNameA(&FileData)==0) return;

	fhand=CreateFile(FileData.lpstrFile,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
	if(fhand==INVALID_HANDLE_VALUE) {
		ReportSystemError("Could not open import file");
		return;
	}

	// determine file type
	switch(FileData.nFilterIndex) {
	case PIFI_PCM:
		p=(PPROJECT)ReadPCMFile(fhand);
		if(p!=NULL) ValidateMesh((PMODEL)p);
	break;
	default: p=NULL;
	}

	CloseHandle(fhand);

	if(p!=NULL) ActivateProject(p);
}
*/

float ReadFloat(HANDLE fhand) {
	DWORD count;
	float f;

	ReadFile(fhand,&f,sizeof(float),&count,NULL);
	if(count<sizeof(float)) EndOfFile=TRUE;

	return f;
}

void WriteFloat(HANDLE fhand,float f) {
	DWORD count;

	WriteFile(fhand,&f,sizeof(float),&count,NULL);
}

long ReadLong(HANDLE fhand) {
	DWORD count;
	long l;

	ReadFile(fhand,&l,sizeof(long),&count,NULL);
	if(count<sizeof(long)) EndOfFile=TRUE;

	return l;
}

void WriteLong(HANDLE fhand,long l) {
	DWORD count;

	WriteFile(fhand,&l,sizeof(long),&count,NULL);
}

short ReadShort(HANDLE fhand) {
	DWORD count;
	short s;

	ReadFile(fhand,&s,sizeof(short),&count,NULL);
	if(count<sizeof(short)) EndOfFile=TRUE;

	return s;
}

void WriteShort(HANDLE fhand,short s) {
	DWORD count;

	WriteFile(fhand,&s,sizeof(short),&count,NULL);
}

void GetBMPFileName(HWND hwnd,int item) {
	FileData.hwndOwner=hwndMain;
	FileData.lpstrFilter=BMPFilter;
	FileData.nFilterIndex=1;
	FileData.lpstrTitle=NULL;
	FileData.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

	if(GetOpenFileNameA(&FileData))
	SetDlgItemText(hwnd,item,FileData.lpstrFile);
}

void WritePaddedString(HANDLE fhand,char *s) {
	DWORD count;

	WriteFile(fhand,s,(strlen(s)+4)&0xffc,&count,NULL);
}

void ReadPaddedString(HANDLE fhand,char *s,int l) {
	DWORD count;
	int ptr;
	BOOL drop;

	ptr=0;
	drop=FALSE;
	do {
		ReadFile(fhand,s+ptr,4,&count,NULL);
		if(s[ptr++]==0) drop=TRUE;
		if(s[ptr++]==0) drop=TRUE;
		if(s[ptr++]==0) drop=TRUE;
		if(s[ptr++]==0) drop=TRUE;
    } while(drop==FALSE && ptr<l);
}

void WriteVecString(HANDLE fhand,float* v) {
	char buff[60];
	sprintf(buff,"<%g,%g,%g>",v[X],v[Y],v[Z]);
	WriteString(fhand,buff);
}

void WriteUVecString(HANDLE fhand,float* u) {
	char buff[40];
	sprintf(buff,"<%g,%g>",u[U],u[V]);
	WriteString(fhand,buff);
}

void WriteString(HANDLE fhand,char* s) {
	DWORD count;

	WriteFile(fhand,s,strlen(s),&count,NULL);
}

void WriteStrings(HANDLE fhand,...) {
	DWORD count;
	va_list ap;
	char *s;

	va_start(ap,fhand);
	do {
		s=va_arg(ap,char*);
		if(s) WriteFile(fhand,s,strlen(s),&count,NULL);
	} while(s);
	va_end(ap);
}

void WriteFloatString(HANDLE fhand,float f) {
	char buff[20];

	sprintf(buff,"%g",f);
	WriteString(fhand,buff);
}

void WriteIntString(HANDLE fhand,int i) {
	char buff[20];

	sprintf(buff,"%d",i);
	WriteString(fhand,buff);
}

void WriteHexString(HANDLE fhand,int i) {
	char buff[20];

	sprintf(buff,"%x",i);
	WriteString(fhand,buff);
}

