#include "data.h"

extern PPROJECT ActiveProject;
extern MYPTR HighItem;
extern short HighType;
extern matrix ibm,iem,qm,bm,em;

knob *CreateKnob(PMODEL m,short pX,short pY,BOOL make_delta) {
	knob *k;

	if(m==NULL) m=(PMODEL)ActiveProject;

	k=(knob*)malloc(sizeof(knob));

	if(k!=NULL) {
		addTail(&m->Knobs,(link*)k);
		mapScr2XYZ(&((PPROJECT)m)->cView,pX,pY,k->Spot);
		k->Flags=0;
		k->Index=m->NextKnobIndex++;
		k->Owner=(part*)NULL;
		((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	}
	return k;
}

PDELT CreateKnobDelta(knob *k,TYPE type) {
	PDELT d;
	PMODEL m;

	m=FindModel((link*)k);

	d=CreateDelta((PPROJECT)m,type,KNOBDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=k->Index;

	if(type==DT_DEL) return d;

	ExtendDelta(d,3,(LSMWORD *)&k->Spot);

	return d;
}

void DestroyKnob(knob *k,BOOL make_delta) {
	PMODEL m;
	part *p,*np;

	if(k) {
		m=FindModel((link*)k);

		STEPPARTS(m,p)
		if(p->Base==k)
		if(hasNonUniChildren(m,p)) {
			ReportSystemError("Cannot delete part.  Part has one or more children that cannot be a base part.");
			return;
		}

		p=FIRSTPART(m);
		while((part*)NEXT(p)) {
			np=(part*)NEXT(p);
			if(p->Base==k || p->End==k || p->Rise==k) DestroyPart(p,make_delta);
			p=np;
		}

		if(make_delta) CreateKnobDelta(k,DT_DEL);

		remLink((link*)k);
		free(k);
		((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	}
	if(HighItem.k==k) {
		HighItem.k=NULL;
		HighType=HT_NOTHING;
	}

	return;
}

void FinishKnobByDelta(PMODEL m,knob *k,PDELT d) {
	k->Index=d->Index;
	k->Spot[X]=d->Data[0].f;
	k->Spot[Y]=d->Data[1].f;
	k->Spot[Z]=d->Data[2].f;
}

// function to move a knob and everything attached to it
void MoveKnobEnd(knob *mk) {
	knob *k;
	vert* v;

	if(mk->Owner) {
		buildMatrices(mk->Owner,(mk->Flags&KF_ISRISE)==0);
		getBaseMatrix(mk,qm);
		MatrixMultiply(ibm,qm,bm);
	}

	getEndMatrix(mk,qm);

	MatrixMultiply(iem,qm,em);

	STEPKNOBS(ActiveProject,k)
	if(k->Flags&TF_DRAGGED)	{
		if(k!=mk) TransformVector(em,k->Save,k->Spot);
	}
	STEPVERTS(ActiveProject,v)
	if(v->Flags&TF_DRAGGED) {
		if((v->Flags&VF_WITHB)!=0 && v->Owner==mk->Owner) {
			TransformVector(bm,v->Save,v->Spot);
		} else {
			TransformVector(em,v->Save,v->Spot);
		}
	}
}

knob *CopyKnob(knob *k,BOOL make_delta) {
	knob *nk;

	nk=CreateKnob(NULL,0,0,make_delta);
	nk->Flags=k->Flags;
	nk->Spot[X]=k->Spot[X];
	nk->Spot[Y]=k->Spot[Y];
	nk->Spot[Z]=k->Spot[Z];
	nk->Owner=NULL;

	return nk;
}

void ReplaceKnob(knob *ok,knob *nk) {
	part* p;

	STEPPARTS(ActiveProject,p)
	if(p->Base==ok) p->Base=nk;
	BaseParts((PMODEL)ActiveProject);
	RebuildPartMatrices();
}

void getEndMatrix(knob* k,float *m) {
	short i;
	if(k->Owner) {
		for(i=0;i<12;i++)
		m[i]=k->Owner->mE[i];
	} else {
		initMatrix(m);
		m[LX]=k->Spot[X];
		m[LY]=k->Spot[Y];
		m[LZ]=k->Spot[Z];
	}
}

void getEndInversion(knob* k,float *m) {
	if(k->Owner) {
		MatrixInvert(k->Owner->mE,m);
	} else {
		initMatrix(m);
		m[LX]=-k->Spot[X];
		m[LY]=-k->Spot[Y];
		m[LZ]=-k->Spot[Z];
	}
}

void getBaseMatrix(knob* k,float *m) {
	register int i;

	for(i=0;i<12;i++)
	m[i]=k->Owner->mB[i];
}

void getBaseInversion(knob* k,float *m) {
	MatrixInvert(k->Owner->mB,m);
}

knob *root(knob* k) {
	knob *rk;

	rk=k;

	while(rk->Owner)
	rk=rk->Owner->Base;

	return rk;
}

