#include <math.h>
#include "data.h"
#include "calc.h"

extern HWND hwndMain,hwndStatusBar;
extern PPROJECT ActiveProject;

matrix IDENTITY={1.0,0.0,0.0, 0.0,1.0,0.0, 0.0,0.0,1.0, 0.0,0.0,0.0 };

float dot(float*v, float* a) {
	return v[X]*a[X]+v[Y]*a[Y]+v[Z]*a[Z];
}

void cross(float*r,float*a,float*b) {
	r[X]=a[Y]*b[Z] - a[Z]*b[Y];
	r[Y]=a[Z]*b[X] - a[X]*b[Z];
	r[Z]=a[X]*b[Y] - a[Y]*b[X];
}

void initMatrix(float* m) {
	m[XX]=m[YY]=m[ZZ]=1.0;
	m[XY]=m[XZ]=m[YX]=
	m[YZ]=m[ZX]=m[ZY]=
	m[LX]=m[LY]=m[LZ]=0.0;
}

void MatrixMultiply(float *a,float *b,float *d) {
	d[XX] =a[XX]*b[XX] + a[XY]*b[YX] + a[XZ]*b[ZX];
	d[XY] =a[XX]*b[XY] + a[XY]*b[YY] + a[XZ]*b[ZY];
	d[XZ] =a[XX]*b[XZ] + a[XY]*b[YZ] + a[XZ]*b[ZZ];

	d[YX] =a[YX]*b[XX] + a[YY]*b[YX] + a[YZ]*b[ZX];
	d[YY] =a[YX]*b[XY] + a[YY]*b[YY] + a[YZ]*b[ZY];
	d[YZ] =a[YX]*b[XZ] + a[YY]*b[YZ] + a[YZ]*b[ZZ];

	d[ZX] =a[ZX]*b[XX] + a[ZY]*b[YX] + a[ZZ]*b[ZX];
	d[ZY] =a[ZX]*b[XY] + a[ZY]*b[YY] + a[ZZ]*b[ZY];
	d[ZZ] =a[ZX]*b[XZ] + a[ZY]*b[YZ] + a[ZZ]*b[ZZ];

	d[LX] =a[LX]*b[XX] + a[LY]*b[YX] + a[LZ]*b[ZX] + b[LX];
	d[LY] =a[LX]*b[XY] + a[LY]*b[YY] + a[LZ]*b[ZY] + b[LY];
	d[LZ] =a[LX]*b[XZ] + a[LY]*b[YZ] + a[LZ]*b[ZZ] + b[LZ];
}

void MatrixInvert(float *s,float *d) {
// not a real inversion; the first nine elements must be
// pure rotation
	d[XX]=s[XX];
	d[XY]=s[YX];
 	d[XZ]=s[ZX];
	d[YX]=s[XY];
 	d[YY]=s[YY];
  	d[YZ]=s[ZY];
	d[ZX]=s[XZ];
 	d[ZY]=s[YZ];
  	d[ZZ]=s[ZZ];

	d[LX]=-dot(&s[XX],&s[LX]);
	d[LY]=-dot(&s[YX],&s[LX]);
	d[LZ]=-dot(&s[ZX],&s[LX]);
}

//void RedoMapMatrices(int cX,int cY) {
//	register short i;
//	RECT wR,sR;
//	view *v;
//	matrix t;

//	if(ActiveProject==NULL) return;

//}

float Normalize(float* s,float* d) {
	float l;
	float* r;

	if((r=d)==NULL) r=s;

	l=(float)sqrt((double)dot(s,s));
	if(l>0.0) {
		r[X]=s[X]/l;
		r[Y]=s[Y]/l;
		r[Z]=s[Z]/l;
	} else {
		r[X]=s[X];
		r[Y]=s[Y];
		r[Z]=s[Z];
	}
	return l;
}

void TransformVector(float *m,float *s,float *d) {
	d[X] = s[X] * m[XX] + s[Y] * m[YX] + s[Z] * m[ZX] + m[LX];
	d[Y] = s[X] * m[XY] + s[Y] * m[YY] + s[Z] * m[ZY] + m[LY];
	d[Z] = s[X] * m[XZ] + s[Y] * m[YZ] + s[Z] * m[ZZ] + m[LZ];
}

void InvertVector(float *m,float *s,float *d) {
	vec src;

	src[X]=s[X]-m[LX];
	src[Y]=s[Y]-m[LY];
	src[Z]=s[Z]-m[LZ];

	d[X]=src[X]*m[XX] + src[Y]*m[XY] + src[Z]*m[XZ];
	d[Y]=src[X]*m[YX] + src[Y]*m[YY] + src[Z]*m[YZ];
	d[Z]=src[X]*m[ZX] + src[Y]*m[ZY] + src[Z]*m[ZZ];
}

void InvertUVector(float *m,float *s,float *d) {
	vec src;

	src[X]=s[X]-m[LX];
	src[Y]=s[Y]-m[LY];
	src[Z]=s[Z]-m[LZ];

	d[X]=src[X]*m[XX] + src[Y]*m[XY] + src[Z]*m[XZ];
	d[Y]=src[X]*m[YX] + src[Y]*m[YY] + src[Z]*m[YZ];
}

float Distance2(float *a, float *b) {
	vec d;

	d[X]=a[X]-b[X];
	d[Y]=a[Y]-b[Y];
	d[Z]=a[Z]-b[Z];
	return dot(d,d);
}

float Distance(float *a, float *b) {
	vec d;

	d[X]=a[X]-b[X];
	d[Y]=a[Y]-b[Y];
	d[Z]=a[Z]-b[Z];
	return (float)sqrt((double)dot(d,d));
}

// polar-to-cartesian conversion
void Polar2Cart(part* p,float *m) {
	float sinx,siny,cosx,cosy,sinz,cosz;
	vec d,e;

	sinx=p->AngleX.use*DEG2RAD;
	siny=p->AngleY.use*DEG2RAD;
	sinz=p->AngleZ.use*DEG2RAD;

	cosx=(float)cos((double)sinx);
	sinx=(float)sin((double)sinx);
	cosy=(float)cos((double)siny);
	siny=(float)sin((double)siny);
	cosz=(float)cos((double)sinz);
	sinz=(float)sin((double)sinz);

	switch(p->Flags&PT_MASK) {
	case PT_UNI:
	case PT_HINGE:
		d[X]=p->Length.use*cosx*siny;
		d[Y]=-p->Length.use*sinx;
		d[Z]=p->Length.use*cosx*cosy;

		TransformVector(m,d,p->End->Spot);

		d[X]=((cosz*sinx*siny - sinz*cosy)*.25      +(cosx*siny)*.75)*p->Length.use;
		d[Y]=((cosz*cosx)*.25                     -(sinx)*.75)*p->Length.use;
		d[Z]=((cosz*sinx*cosy + sinz*siny)*.25      +(cosx*cosy)*.75)*p->Length.use;
	break;
	case PT_BALL:
		d[X]=p->Length.use*siny*sinx;
		d[Y]=-p->Length.use*cosy*sinx;
		d[Z]=p->Length.use*cosx;

		TransformVector(m,d,p->End->Spot);

		e[X]=-sinz*.25*p->Length.use;
		e[Y]= cosz*.25*p->Length.use;
		e[Z]= .75*p->Length.use;

		d[X]=e[X]*cosy+e[Y]*siny;
		d[Y]=e[Y]*cosy-e[X]*siny;
		d[Z]=e[Z];

		e[X]=d[X];
		e[Y]=d[Y]*cosx-d[Z]*sinx;
		e[Z]=d[Z]*cosx+d[Y]*sinx;

		d[X]=e[X]*cosy-e[Y]*siny;
		d[Y]=e[Y]*cosy+e[X]*siny;
		d[Z]=e[Z];
	break;
	}

	TransformVector(m,d,p->Rise->Spot);
	p->TwistCos=cosz;
	p->TwistSin=sinz;
}

// cartesian-to-polar conversion
void Cart2Polar(part* p,float *m) {
	float dist;
	vec d,l;

	d[X]=p->End->Spot[X]-m[LX];
	d[Y]=p->End->Spot[Y]-m[LY];
	d[Z]=p->End->Spot[Z]-m[LZ];

	p->Length.tmp=dist=(float)sqrt((double)dot(d,d));

	l[X]=dot(d,&m[XX]);
	l[Y]=dot(d,&m[YX]);
	l[Z]=dot(d,&m[ZX]);

	switch(p->Flags&PT_MASK) {
	case PT_UNI:
		if(dist==0.0) {
			p->AngleX.tmp=0.0;
		} else {
			p->AngleX.tmp=-RAD2DEG*(float)asin((double)(l[Y]/dist));
		}
		if(l[X]==0.0 && l[Z]==0.0) {
			p->AngleY.tmp=0.0;
		} else {
			p->AngleY.tmp=RAD2DEG*(float)atan2((double)l[X],(double)l[Z]);
		}
	break;
	case PT_HINGE:
		if(dist==0.0) {
			p->AngleX.tmp=0.0;
		} else {
			p->AngleX.tmp=-RAD2DEG*(float)atan2((double)l[Y],(double)l[Z]);
		}
		p->AngleY.tmp=0.0;
	break;
	case PT_BALL:
		if(dist==0.0) {
			p->AngleX.tmp=0.0;
		} else {
			p->AngleX.tmp=RAD2DEG*(float)acos((double)l[Z]/dist);
		}
		if(p->AngleX.tmp==0.0) {
			p->AngleY.tmp=0.0;
		} else {
			p->AngleY.tmp=RAD2DEG*(float)atan2((double)l[X],(double)-l[Y]);
		}
	break;
	}

	p->AngleZ.tmp=RAD2DEG*(float)atan2((double)p->TwistSin,(double)p->TwistCos);
}

void SubtractVector(float *a,float *b,float *d) {

	if(d==NULL) d=a;

	d[X]=a[X]-b[X];
	d[Y]=a[Y]-b[Y];
	d[Z]=a[Z]-b[Z];
}

void CalcModelAngles(PMODEL m) {
	MYPTR mp;
	matrix ma;

	if(m==NULL) return;

	STEPPARTS(m,mp.p) {
		getInheritedMatrix(mp.p,ma);
		Cart2Polar(mp.p,ma);
	}
}

BOOL isPowerof2(UINT value) {
	while(value>1) {
		if(value&1) return FALSE;
		value>>=1;
	}

	return TRUE;
}

UINT nearestPowerof2(UINT value) {
	UINT pow;

	value=(value<<8)/181;

	pow=1;
	while (value>1) {
		value>>=1;
		pow<<=1;
	}
	return pow;
}

void MapXYZ2UV(float *s,float *d) {
	float fac;
	InvertUVector(ActiveProject->cView.Orient,s,d);
	fac=ActiveProject->cView.Scale;

	d[U]/=(fac*2.0);d[U]+=.5;
	d[V]/=(fac*2.0);d[V]+=.5;
//	d.z=0.0;
}

