#include "data.h"
#include <math.h>

extern HWND hwndMain;
extern PPROJECT ActiveProject;
extern vec scr,xyz,tmp;
extern short HighType;
extern MYPTR HighItem;

part* CreatePart(PMODEL m,knob *b,knob *e,knob *r,BOOL make_delta) {
	part* p;
//	float len;

	if(b==NULL) ReportSystemError("Knob b==NULL");

	if(m==NULL) m=(PMODEL)ActiveProject;

	p=(part*)malloc(sizeof(part));
	if(p!=NULL) {
		addTail(&m->Parts,(link*)p);
		p->Base=b;
		p->End=e;
		p->Rise=r;
		p->Flags=0;
		p->Index=m->NextPartIndex++;
		p->Parent=b->Owner;
		e->Owner=p;
		p->AngleX.use=p->AngleY.use=p->AngleZ.use=0.0;
		p->TwistCos=1.0;
		p->TwistSin=0.0;
		p->Name=NULL;
		p->Name=CreatePartName(m);
		buildMatrices(p,r==NULL);
		p->Rise->Flags|=KF_ISRISE;
		p->Rise->Owner=p;
//		((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	}
	return p;
}

PDELT CreatePartDelta(part* p,TYPE type) {
	PDELT d;
	PMODEL m;
	LSMWORD extra[4];

	m=FindModel((link*)p);

	d=CreateDelta((PPROJECT)m,type,PARTDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=p->Index;
	if(type==DT_DEL) return d;

	extra[0].l=p->Base->Index;
	extra[1].l=p->End->Index;
	extra[2].l=p->Rise->Index;
	ExtendDelta(d,3,extra);

// flags
	if((p->Flags&PT_MASK)==PT_BALL) {
		extra[0].l=BALLFLAG;
		ExtendDelta(d,1,extra);
	}
	if((p->Flags&PT_MASK)==PT_HINGE) {
		extra[0].l=HINGEFLAG;
		ExtendDelta(d,1,extra);
	}
	if(p->Flags&PF_LOCKX) {
		extra[0].l=LOCKXBLOCK;
		ExtendDelta(d,1,extra);
	}
	if(p->Flags&PF_LOCKY) {
		extra[0].l=LOCKYBLOCK;
		ExtendDelta(d,1,extra);
	}
	if(p->Flags&PF_LOCKZ) {
		extra[0].l=LOCKZBLOCK;
		ExtendDelta(d,1,extra);
	}
	if(p->Flags&PF_LOCKL) {
		extra[0].l=LOCKLBLOCK;
		ExtendDelta(d,1,extra);
	}
	if(p->Name!=NULL) {
		ExtendDelta(d,(strlen(p->Name)+4)>>2,(LSMWORD*)p->Name);
	}
	return d;
}

// build matrices
void buildMatrices(part* p,BOOL fordrag) {
//	short i;
	knob *b,*e,*r;
	PMODEL m;
	float fX,fY,fZ;
	matrix t;

//	if(p==NULL) ReportSystemError("NULL pointer passed to BuildMatrices()");

	m=FindModel((link*)p);

	getInheritedMatrix(p,t);

	b=p->Base;
	e=p->End;
	r=p->Rise;

// put location vectors in place
	p->mB[LX]=b->Spot[X];
	p->mB[LY]=b->Spot[Y];
	p->mB[LZ]=b->Spot[Z];
	p->mE[LX]=e->Spot[X];
	p->mE[LY]=e->Spot[Y];
	p->mE[LZ]=e->Spot[Z];

// get starting z
	SubtractVector(&p->mE[LX],&p->mB[LX],&p->mB[ZX]);

	p->Length.use=Normalize(&p->mB[ZX],NULL);

// end z vector is equal to start z vector
	p->mE[ZX]=p->mB[ZX];
	p->mE[ZY]=p->mB[ZY];
	p->mE[ZZ]=p->mB[ZZ];

	switch(p->Flags&PT_MASK) {
		case PT_UNI:
			cross(&p->mB[XX],&t[YX],&p->mB[ZX]);
			if(Normalize(&p->mB[XX],NULL)==0.0) {
				p->mB[XX]=t[XX];
				p->mB[XY]=t[XY];
				p->mB[XZ]=t[XZ];
			}
		break;
		case PT_HINGE:
			p->mB[XX]=t[XX];
			p->mB[XY]=t[XY];
			p->mB[XZ]=t[XZ];
		break;
		case PT_BALL:
			fX=dot(&t[XX],&p->mB[ZX]);
			fY=dot(&t[YX],&p->mB[ZX]);
			fZ=dot(&t[ZX],&p->mB[ZX]);
			if(fZ==1.0) {
				p->mB[XX]=t[XX];
				p->mB[XY]=t[XY];
				p->mB[XZ]=t[XZ];
			}
			else if(fZ==-1.0) {
				p->mB[XX]=-t[XX];
				p->mB[XY]=-t[XY];
				p->mB[XZ]=-t[XZ];
			}
			else {
				cross(&p->mB[YX],&t[ZX],&p->mB[ZX]);
				Normalize(&p->mB[YX],NULL);
				cross(&p->mB[XX],&p->mB[YX],&t[ZX]);
				fX=dot(&t[XX],&p->mB[XX]);
				fY=dot(&t[XX],&p->mB[YX]);
				cross(&p->mB[XX],&p->mB[YX],&p->mB[ZX]);
				p->mB[XX]=fX*p->mB[XX]+fY*p->mB[YX];
				p->mB[XY]=fX*p->mB[XY]+fY*p->mB[YY];
				p->mB[XZ]=fX*p->mB[XZ]+fY*p->mB[YZ];
			}
		break;
	} // end of switch
	cross(&p->mB[YX],&p->mB[ZX],&p->mB[XX]);

// make Rise, if needed
	if(r==NULL) { // no riser defined yet
		p->Rise=r=CreateKnob(m,0,0,TRUE);
		fordrag=TRUE;
	}

// build end matrix
	if(fordrag) { // riser is not valid
		p->mE[XX]=p->mB[XX]*p->TwistCos + p->mB[YX]*p->TwistSin;
		p->mE[XY]=p->mB[XY]*p->TwistCos + p->mB[YY]*p->TwistSin;
		p->mE[XZ]=p->mB[XZ]*p->TwistCos + p->mB[YZ]*p->TwistSin;

		p->mE[YX]=p->mB[YX]*p->TwistCos - p->mB[XX]*p->TwistSin;
		p->mE[YY]=p->mB[YY]*p->TwistCos - p->mB[XY]*p->TwistSin;
		p->mE[YZ]=p->mB[YZ]*p->TwistCos - p->mB[XZ]*p->TwistSin;

		CalcRise(p);
	} else { // riser is valid
		p->mE[YX]=r->Spot[X]-b->Spot[X];
		p->mE[YY]=r->Spot[Y]-b->Spot[Y];
		p->mE[YZ]=r->Spot[Z]-b->Spot[Z];
		cross(&p->mE[XX],&p->mE[YX],&p->mE[ZX]);
		Normalize(&p->mE[XX],NULL);
		cross(&p->mE[YX],&p->mE[ZX],&p->mE[XX]);
		Normalize(&p->mE[YX],NULL);
		p->TwistCos=dot(&p->mE[XX],&p->mB[XX]);
		p->TwistSin=dot(&p->mE[XX],&p->mB[YX]);
	}
}

void DestroyPart(part* p,BOOL make_delta) {
	PMODEL m;
	vert* v;
	part* p1;
	knob *k;

	if(p==NULL) return;

	m=FindModel((link*)p);
	STEPVERTS(m,v)
	if(v->Owner==p)
	DetachVertex(v,make_delta);

	STEPPARTS(m,p1)
	if(p1->Parent==p) UnleashPart(p1,make_delta);

	STEPKNOBS(m,k)
	if(k->Owner==p) {
		k->Owner=NULL;
		if(make_delta) CreateKnobDelta(k,DT_CHG);
	}

	p->Base->Owner=p->Parent;

	if(HighItem.p==p) {
		HighItem.p=NULL;
		HighType=HT_NOTHING;
	}

	k=p->Rise;

	k->Owner=(part*)0;

	if(make_delta) CreatePartDelta(p,DT_DEL);

	((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	remLink((link*)p);
	DestroyString(p->Name);
	free(p);

	DestroyKnob(k,make_delta);
	return;
}

void FinishPartByDelta(PMODEL m,part* p,PDELT d) {
	UINT i;

	p->Base=IndexKnob(m,d->Data[0].l);
	p->End=IndexKnob(m,d->Data[1].l);
	p->Rise=IndexKnob(m,d->Data[2].l);

	i=3;
	while(i<d->Len)
	switch(d->Data[i].l) {
		case BALLFLAG:
			p->Flags&=~PT_MASK;
			p->Flags|=PT_BALL;
			i++;
		break;
		case HINGEFLAG:
			p->Flags&=~PT_MASK;
			p->Flags|=PT_HINGE;
			i++;
		break;
		case LIMXBLOCK:	i+=3;	break;	// no longer supported
		case LIMYBLOCK:	i+=3;	break;	// no longer supported
		case LIMZBLOCK:	i+=3;	break;	// no longer supported
		case LIMLBLOCK:	i+=3;	break;	// no longer supported
		case LOCKXBLOCK:	p->Flags|=PF_LOCKX;	i++;	break;
		case LOCKYBLOCK:	p->Flags|=PF_LOCKY;	i++;	break;
		case LOCKZBLOCK:	p->Flags|=PF_LOCKZ;	i++;	break;
		case LOCKLBLOCK:	p->Flags|=PF_LOCKL;	i++;	break;
		default: i++;
	}
}

void UnleashPart(part* p,BOOL make_delta) {
	if(p!=NULL) {
		p->Parent=NULL;
		p->Base->Owner=NULL;
		if(make_delta) CreatePartDelta(p,DT_CHG);
	}
}

// This is to get all matrices rehashed after a knob dragging operation
void RebuildPartMatrices(void) {
	BOOL hit;
	part* p;

	do {
		hit=FALSE;
		STEPPARTS(ActiveProject,p)
		if((p->Flags&TF_DRAGGED)!=0)
		if(p->Parent==NULL) {
//			if(p==NULL) ReportSystemError("Bogey!");
			buildMatrices(p,FALSE);
			p->Flags&=~TF_DRAGGED;
			hit=TRUE;
		} else {
			if((p->Parent->Flags&TF_DRAGGED)==0) {
//				if(p==NULL) ReportSystemError("Bogey!");
				buildMatrices(p,FALSE);
				p->Flags&=~TF_DRAGGED;
				hit=TRUE;
			}
		}
	} while(hit!=FALSE);
}

void getInheritedMatrix(part* p,float* m) {
	register short i;
//	matrix s;

	initMatrix(m);

	if(p->Parent!=NULL)
	for(i=0;i<9;i++)
	m[i]=p->Parent->mE[i];

	m[LX]=p->Base->Spot[X];
	m[LY]=p->Base->Spot[Y];
	m[LZ]=p->Base->Spot[Z];
}

void RebuildAllMatrices(PMODEL m) {
	BOOL hit;
	part* p;

	if(m==NULL) m=(PMODEL)ActiveProject;
	if(m==NULL) return;

	STEPPARTS(m,p) p->Flags&=~TF_DRAGGED;

	do {
		hit=FALSE;
		STEPPARTS(m,p)
		if((p->Flags&TF_DRAGGED)==0)
		if(p->Parent==NULL) {
			buildMatrices(p,FALSE);
			p->Flags|=TF_DRAGGED;
			hit=TRUE;
		} else {
			if((p->Parent->Flags&TF_DRAGGED)!=0) {
				buildMatrices(p,FALSE);
				p->Flags|=TF_DRAGGED;
				hit=TRUE;
			}
		}
	} while(hit!=FALSE);
}

void BaseParts(PMODEL m) {
	part* p;

	STEPPARTS(m,p)
	p->Parent=p->Base->Owner;
}

/*
void BuildBaseMatrix(part* p) {
	MATRIX m;

	GetInheritedMatrix(p,m);

	SubtractVector(p->End->Spot,p->Base->Spot,&p->mB[ZX]);
	Normalize(&m[ZX],NULL);

	CrossProduct(&m[YX],&p->mB[ZX],&p->mB[XX]);
	Normalize(&p->mB[XX],NULL);
	CrossProduct(&p->mB[ZX],&p->mB[XX],&p->mB[YX]);
	p->mB[LX]=p->Base->Spot[X];
	p->mB[LY]=p->Base->Spot[Y];
	p->mB[LZ]=p->Base->Spot[Z];
}

void BuildEndMatrix(part* p) {

	p->mE[ZX]=p->mB[ZX];
	p->mE[ZY]=p->mB[ZY];
	p->mE[ZZ]=p->mB[ZZ];
	p->mE[LX]=p->End->Spot[X];
	p->mE[LY]=p->End->Spot[Y];
	p->mE[LZ]=p->End->Spot[Z];

	if(p->Rise==NULL) {
		p->Rise=CreateKnob(NULL,0,0);
		p->Flags|=PF_BADRISE;
	}

	if(f->Flgas&PF_BADRISE) {


	} else {
		SubtractVector(

*/

void CalcRise(part* p) {
	float dp;

	dp=Distance(p->End->Spot,p->Base->Spot)*.25;

	p->Rise->Spot[X]=p->End->Spot[X]*.75 + p->Base->Spot[X]*.25 + dp*p->mE[YX];
	p->Rise->Spot[Y]=p->End->Spot[Y]*.75 + p->Base->Spot[Y]*.25 + dp*p->mE[YY];
	p->Rise->Spot[Z]=p->End->Spot[Z]*.75 + p->Base->Spot[Z]*.25 + dp*p->mE[YZ];
}

void Constrain(knob *k) {
	part* p;
	vec q1;
	float dp;
	matrix qm;

	if(k==NULL) return;

	p=k->Owner;
	if(p==NULL) return;

	if(k->Flags&KF_ISRISE) {
		q1[X]=k->Spot[X]-p->Base->Spot[X];
		q1[Y]=k->Spot[Y]-p->Base->Spot[Y];
		q1[Z]=k->Spot[Z]-p->Base->Spot[Z];

		dp=dot(q1,&p->mB[ZX]);
		q1[X]-=dp*p->mB[ZX];
		q1[Y]-=dp*p->mB[ZY];
		q1[Z]-=dp*p->mB[ZZ];
		Normalize(q1,NULL);
		p->TwistCos=dot(q1,&p->mB[YX]);
		p->TwistSin=-dot(q1,&p->mB[XX]);

		CalcRise(p);
	}

	getInheritedMatrix(p,qm);
	Cart2Polar(p,qm);

	if((p->Flags&PF_LOCKX)==0) p->AngleX.use=p->AngleX.tmp;
	if((p->Flags&PF_LOCKY)==0) p->AngleY.use=p->AngleY.tmp;
	if((p->Flags&PF_LOCKZ)==0) p->AngleZ.use=p->AngleZ.tmp;
	if((p->Flags&PF_LOCKL)==0) p->Length.use=p->Length.tmp;

	Polar2Cart(p,qm);
}

void ConvertJoint(part* p,TYPE new_type) {
//	matrix qm;

	if((p->Flags&PT_MASK)==new_type) return;

	if(p->Parent==NULL)
	if(new_type==PT_BALL || new_type==PT_HINGE) {
		ReportSystemError("Only child parts can be made into this joint type.");
		return;
	}

	PrepSingleKnobDrag(p->End);
//	getInheritedMatrix(p,qm);

	p->Flags&=~PT_MASK;
	p->Flags|=new_type;

//	Cart2Polar(p,qm);
//	Polar2Cart(p,qm);
	MoveKnobEnd(p->End);
	RebuildPartMatrices();
}

BOOL hasNonUniChildren(PMODEL m,part* p) {
	part *op;

	STEPPARTS(m,op)
	if(op->Parent==p)
	if((op->Flags&PT_MASK)!=PT_UNI)
	return TRUE;

	return FALSE;
}

