#include <windows.h>
#include <gl/gl.h>
#include "data.h"

extern chain Projects;
extern float IDENTITY[12];
extern HWND hwndMain;
extern HMENU ModelMenu,LameMenu;
//extern MYPTR HighItem;
//extern short HighType;

PPROJECT ActiveProject=(PPROJECT)NULL;

PPROJECT CreateProject(void /*char *Name*/) {
	PPROJECT p;
//	register short i;

	p=(PPROJECT)malloc(sizeof(PROJECT));

	if(p!=(PPROJECT)NULL) InitProject(p);
	return p;
}

void InitProject(PPROJECT p) {

	if(p!=(PPROJECT)NULL) {
		memset(p,0,sizeof(PROJECT));
		addHead(&Projects,&p->Link);
		p->Flags=PF_SHOWAXES|PF_INVALIDVIEW;
		initView(&p->mView);
		initView(&p->cView);
		initView(&p->uView);
		p->EditMode=EM_MESHEDIT;
//		p->SaveTitle=NULL;
//		p->SavePath=NULL;
//		p->ExportTitle=NULL;
//		p->ExportPath=NULL;
		p->UserMatrix[XX]=p->UserMatrix[YY]=p->UserMatrix[ZZ]=
		p->UserScale[X]=p->UserScale[Y]=p->UserScale[Z]=1.0;
		initChain(&p->Deltas);
//		p->CurrentDelta=NULL;
		CreateDelta(p,DT_ATOM,0,0);
	}
}

BOOL DestroyProject(PPROJECT p) {

	if(p==NULL) return TRUE;
	if(p->Flags&PF_MODIFIED) {
		ActivateProject(p);
		if(MessageBoxEx(hwndMain,"This project has been modified.  Save it?\n","Shutting down",
			MB_YESNOCANCEL,LANG_NEUTRAL)==IDCANCEL) return FALSE;
	}
	if(p->ProjectType==PT_MODEL) DestroyModelInfo((PMODEL)p);
	if(p->SaveTitle) DestroyString(p->SaveTitle);
	if(p->SavePath) DestroyString(p->SavePath);
	if(p->ExportTitle) DestroyString(p->ExportTitle);
	if(p->ExportPath) DestroyString(p->ExportPath);
	DestroyAllDeltas(p);
	remLink(&p->Link);
	free(p);
	return TRUE;
}

void DestroyAllDeltas(PPROJECT p) {
	while(p->Deltas.pop) DestroyDelta(FIRSTDELT(p));
	p->CurrentDelta=NULL;
}

void ActivateNextProject(void) {
	PPROJECT p;

	SaveView();

	if(Projects.pop) {
		p=(PPROJECT)Projects.Next;
		remLink(&p->Link);
		addTail(&Projects,&p->Link);
	}

	ActivateFirstProject();
}

void ActivatePrevProject(void) {
	PPROJECT p;

	SaveView();

	if(Projects.pop) {
		p=(PPROJECT)Projects.Stub.Prev;
		remLink(&p->Link);
		addHead(&Projects,&p->Link);
	}

	ActivateFirstProject();
}

void ActivateFirstProject(void) {

	SaveView();

	if(Projects.pop) {
		ActivateProject((PPROJECT)Projects.Next);
	} else {
		ActivateProject(NULL);
	}
}

void ActivateProject(PPROJECT p) {

	SaveView();

	ActiveProject=p;

	if(p==NULL) {
		SetMenu(hwndMain,LameMenu);
	} else {
		remLink(&p->Link);
		addHead(&Projects,&p->Link);
		SetMenu(hwndMain,ModelMenu);
		ClearHighlight();
		ClearAll();
	}

	SetWindowTitle();
	UpdateModelMenu();
	UpdateToolBar();
	InvalidateView();
}

BOOL DestroyAllProjects(void) {
	while(Projects.pop!=0) {
		if(DestroyProject((PPROJECT)(Projects.Next))==FALSE)
		return FALSE;
	}
	return TRUE;
}

void CloseActiveProject(void) {
	if(ActiveProject) DestroyProject(ActiveProject);

	ActivateFirstProject();
}


void EnterMode(short newMode) {
	if(ActiveProject==NULL) return;

	SaveView();
	ActiveProject->EditMode=newMode;

	switch(newMode) {
		case EM_MESHEDIT:
			ActiveProject->cView=ActiveProject->mView;
		break;
		case EM_MAPEDIT:
			ActiveProject->cView=ActiveProject->uView;
			ViewXY();
		break;
	}

	ClearHighlight();
	ClearAll();
	InvalidateView();
	SetMouseMode(MM_NOTHING);
}

void SaveView(void) {
	if(ActiveProject==NULL) return;

	switch(ActiveProject->EditMode) {
		case EM_MESHEDIT:
			ActiveProject->mView=ActiveProject->cView;
		break;
		case EM_MAPEDIT:
			ActiveProject->uView=ActiveProject->cView;
		break;
	}
}

PDELT PreviousAtomDelta(PPROJECT p) {
	PDELT d;

	d=(PDELT)(p->CurrentDelta->Link.Prev);

	while(d->Link.Prev!=NULL) {
		if(d->DeltaType==DT_ATOM) return d;
		d=(PDELT)d->Link.Prev;
	}

	return NULL;
}

PDELT NextAtomDelta(PPROJECT p) {
	PDELT d;

	d=(PDELT)(p->CurrentDelta->Link.Next);

	while(d->Link.Next!=NULL) {
		if(d->DeltaType==DT_ATOM) return d;
		d=(PDELT)d->Link.Next;
	}

	return NULL;
}

