// undo.c

#include "data.h"

void Undo(PPROJECT p) {
	PDELT d;
	PDELT od;

	if(p==NULL) return;

	if(PreviousAtomDelta(p)==NULL) return; // no undoing to do!

	d=p->CurrentDelta;
	do {
		d=(PDELT)(d->Link.Prev);
		if(d->Link.Prev==NULL) {
			ReportSystemError("Error tracing backward in delta chain");
			return;
		}

		switch(d->DeltaType) {
			case DT_ADD:
				RemoveObjectByDelta(p,d);
			break;
			case DT_CHG:
				od=PreviousMatch(d);
				if(od!=NULL)
				ReplaceObjectByDelta(p,od);
			break;
			case DT_DEL:
				od=PreviousMatch(d);
				if(od!=NULL)
				RebuildObjectByDelta(p,od);
			break;
//			case DT_ATOM:
//				ReportSystemError("DT_ATOM reached in undo");
//			break;
		}
	} while (d->DeltaType!=DT_ATOM);

	p->CurrentDelta=d;
	ValidateMesh((PMODEL)p);
	InvalidateView();
}

void Redo(PPROJECT p) {
	PDELT d;

	if(p==NULL) return;

	if(NextAtomDelta(p)==NULL) return; // no redoing to do!

	d=p->CurrentDelta;
	do {
		d=(PDELT)(d->Link.Next);
		if(d->Link.Next==NULL) {
			ReportSystemError("Error tracing forward in delta chain");
			return;
		}

		switch(d->DeltaType) {
			case DT_ADD:
				RebuildObjectByDelta(p,d);
			break;
			case DT_CHG:
				ReplaceObjectByDelta(p,d);
			break;
			case DT_DEL:
				RemoveObjectByDelta(p,d);
			break;
		}
	} while (d->DeltaType!=DT_ATOM);

	p->CurrentDelta=d;
	ValidateMesh((PMODEL)p);
	InvalidateView();
}

PDELT NextMatch(PDELT d) {
	PDELT match;

	for(match=(PDELT)d->Link.Next;match->Link.Next!=NULL;match=(PDELT)match->Link.Next) {
		if(match->Index==d->Index && match->Data[0].l==d->Data[0].l) return match;
	}
	return NULL;
}

PDELT PreviousMatch(PDELT d) {
	PDELT match;

	for(match=(PDELT)d->Link.Prev;match->Link.Prev!=NULL;match=(PDELT)match->Link.Prev) {
		if(match->Index==d->Index && match->Data[0].l==d->Data[0].l) return match;
	}
	return NULL;
}

