#include "data.h"
#include <math.h>

extern PPROJECT ActiveProject;

void AddMapping(PMODEL m,face* f,BOOL pull) {

	if(m==NULL) m=(PMODEL)ActiveProject;
	if(m==NULL) return;
	if(f==NULL) return;

	f->MapA = f->MapB = f->MapC = f->MapD = NULL;
	if(pull) {
		PullUVData(f,f->EdgeE);
		PullUVData(f,f->EdgeF);
		PullUVData(f,f->EdgeG);
		if(f->VertD) PullUVData(f,f->EdgeH);
	}
	if(f->MapA==NULL || f->MapB==NULL || f->MapC==NULL || (f->MapD==NULL && f->VertD!=NULL) )
	CompleteUVData(f);
	CreateFaceDelta(f,DT_CHG);
}

void CompleteUVData(face* f) {
	uvec uv;
	short flags;

	flags=0;
	if(f->MapA) { flags=1; } else { f->MapA=CreateUVPt(NULL,TRUE); }
	if(f->MapB) { flags|=2; } else { f->MapB=CreateUVPt(NULL,TRUE); }
	if(f->MapC) { flags|=4; } else { f->MapC=CreateUVPt(NULL,TRUE); }
	if(f->VertD) {
		if(f->MapD) { flags|=8; } else { f->MapD=CreateUVPt(NULL,TRUE); }
		switch(flags) {
			case  0: // no bordering uv data to duplicate
				MapXYZ2UV(f->VertA->Spot,f->MapA->Spot);
				MapXYZ2UV(f->VertB->Spot,f->MapB->Spot);
				MapXYZ2UV(f->VertC->Spot,f->MapC->Spot);
				MapXYZ2UV(f->VertD->Spot,f->MapD->Spot);
			break;
//			case  1: break; // should not arise
//			case  2: break; // should not arise
			case  3: // build uv coordinates for C and D
				ExtendMapping(f,f->EdgeE,uv);
				f->MapC->Spot[U] = f->MapB->Spot[U] + uv[U];
				f->MapC->Spot[V] = f->MapB->Spot[V] + uv[V];
				f->MapD->Spot[U] = f->MapA->Spot[U] + uv[U];
				f->MapD->Spot[V] = f->MapA->Spot[V] + uv[V];
			break;
//			case  4: break; // should not arise
//			case  5: break; // should not arise
			case  6: // build uv coordinates for A and D
				ExtendMapping(f,f->EdgeF,uv);
				f->MapD->Spot[U] = f->MapC->Spot[U] + uv[U];
				f->MapD->Spot[V] = f->MapC->Spot[V] + uv[V];
				f->MapA->Spot[U] = f->MapB->Spot[U] + uv[U];
				f->MapA->Spot[V] = f->MapB->Spot[V] + uv[V];
			break;
			case  7: // build uv coordinates for D
				f->MapD->Spot[U] = f->MapA->Spot[U] + f->MapC->Spot[U] - f->MapB->Spot[U];
				f->MapD->Spot[V] = f->MapA->Spot[V] + f->MapC->Spot[V] - f->MapB->Spot[V];
			break;
//			case  8: break; // should not arise
			case  9: // build uv coordinates for B and C
				ExtendMapping(f,f->EdgeH,uv);
				f->MapB->Spot[U] = f->MapA->Spot[U] + uv[U];
				f->MapB->Spot[V] = f->MapA->Spot[V] + uv[V];
				f->MapC->Spot[U] = f->MapD->Spot[U] + uv[U];
				f->MapC->Spot[V] = f->MapD->Spot[V] + uv[V];
			break;
//			case 10: break; // should not arise
			case 11: // build uv coordinates for C
				f->MapC->Spot[U] = f->MapB->Spot[U] + f->MapD->Spot[U] - f->MapA->Spot[U];
				f->MapC->Spot[V] = f->MapB->Spot[V] + f->MapD->Spot[V] - f->MapA->Spot[V];
			break;
			case 12: // build uv coordinates for A and B
				ExtendMapping(f,f->EdgeG,uv);
				f->MapA->Spot[U] = f->MapD->Spot[U] + uv[U];
				f->MapA->Spot[V] = f->MapD->Spot[V] + uv[V];
				f->MapB->Spot[U] = f->MapC->Spot[U] + uv[U];
				f->MapB->Spot[V] = f->MapC->Spot[V] + uv[V];
			break;
			case 13: // build uv coordinates for B
				f->MapB->Spot[U] = f->MapA->Spot[U] + f->MapC->Spot[U] - f->MapD->Spot[U];
				f->MapB->Spot[V] = f->MapA->Spot[V] + f->MapC->Spot[V] - f->MapD->Spot[V];
			break;
			case 14: // build uv coordinates for A
				f->MapA->Spot[U] = f->MapB->Spot[U] + f->MapD->Spot[U] - f->MapC->Spot[U];
				f->MapA->Spot[V] = f->MapB->Spot[V] + f->MapD->Spot[V] - f->MapC->Spot[V];
			break;
//			case 15: break; // no action needed
		}
	} else {
		switch(flags) {
			case 0: // need all three
				MapXYZ2UV(f->VertA->Spot,f->MapA->Spot);
				MapXYZ2UV(f->VertB->Spot,f->MapB->Spot);
				MapXYZ2UV(f->VertC->Spot,f->MapC->Spot);
			break;
//			case 1: // need B and C; should not arise
//			break;
//			case 2: // need A and C; should not arise
//			break;
			case 3: // need C
				ExtendMapping(f,f->EdgeE,uv);
				f->MapC->Spot[U] = (f->MapA->Spot[U] + f->MapB->Spot[U])*.5
					+ uv[U]*SQRT_3_4THS;
				f->MapC->Spot[V] = (f->MapA->Spot[V] + f->MapB->Spot[V])*.5
					+ uv[V]*SQRT_3_4THS;
			break;
//			case 4: // need A and B; should not arise
//			break;
			case 5: // need B
				ExtendMapping(f,f->EdgeG,uv);
				f->MapB->Spot[U] = (f->MapC->Spot[U] + f->MapA->Spot[U])*.5
					+ uv[U]*SQRT_3_4THS;
				f->MapB->Spot[V] = (f->MapC->Spot[V] + f->MapA->Spot[V])*.5
					+ uv[V]*SQRT_3_4THS;
			break;
			case 6: // need A
				ExtendMapping(f,f->EdgeF,uv);
				f->MapA->Spot[U] = (f->MapB->Spot[U] + f->MapC->Spot[U])*.5
					+ uv[U]*SQRT_3_4THS;
				f->MapA->Spot[V] = (f->MapB->Spot[V] + f->MapC->Spot[V])*.5
					+ uv[V]*SQRT_3_4THS;
			break;
//			case 7: // all done; no action required
//			break;
		}
	}
}

uvpt *CopyUVPt(uvpt* u) {
	uvpt *c;

	if(u==NULL) return NULL;

	c=CreateUVPt(NULL,TRUE);
	if(c) {
		c->Spot[U]=u->Spot[U];
		c->Spot[V]=u->Spot[V];
	}
	return c;
}

uvpt* CreateUVPt(PMODEL m,BOOL make_delta) {
	uvpt* u;

	if(m==NULL) m=(PMODEL)ActiveProject;
	if(m==NULL) return NULL;

	u=(uvpt*)malloc(sizeof(uvpt));
	if(u) {
		addTail(&m->UVPts,(link*)u);
		u->Flags=0;
		u->Index=m->NextUVPtIndex++;
		if(make_delta) CreateUVPtDelta(u,DT_ADD);
	}
	return u;
}

PDELT CreateUVPtDelta(uvpt* u,TYPE type) {
	PDELT d;
	PMODEL m;

	m=FindModel((link*)u);

	d=CreateDelta((PPROJECT)m,type,UVPTDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=u->Index;
	if(type==DT_DEL) return d;

	ExtendDelta(d,2,(LSMWORD *)&u->Spot);

	return d;
}

void DestroyUVPt(uvpt* u,BOOL make_delta) {
	PMODEL m;

	m=FindModel((link*)u);

	if(make_delta) CreateUVPtDelta(u,DT_DEL);

	if(m) ((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	remLink((link*)u);
	free(u);
}

void ExtendMapping(face* f,edge* e,float *uv) {
	uvpt *uo,*up,*un;
	uvec temp;
	face* of;

	if(f==NULL) return;
	if(e==NULL) return;
	if(uv==NULL) return;
	of=OtherSide(e,f);
	if(of==NULL) return;

	if(of->VertD==NULL) { // triangle
		if(e==of->EdgeE) { uo=of->MapA; up=of->MapB; un=of->MapC; }
		if(e==of->EdgeF) { uo=of->MapB; up=of->MapC; un=of->MapA; }
		if(e==of->EdgeG) { uo=of->MapC; up=of->MapA; un=of->MapB; }
	} else { //quadrilateral
		if(e==of->EdgeE) { uo=of->MapA; up=of->MapB; un=of->MapD; }
		if(e==of->EdgeF) { uo=of->MapB; up=of->MapC; un=of->MapA; }
		if(e==of->EdgeG) { uo=of->MapC; up=of->MapD; un=of->MapB; }
		if(e==of->EdgeH) { uo=of->MapD; up=of->MapA; un=of->MapC; }
	}
	uv[U]=up->Spot[V]-uo->Spot[V];
	uv[V]=uo->Spot[U]-up->Spot[U];
    temp[U]=un->Spot[U]-uo->Spot[U];
    temp[V]=un->Spot[V]-uo->Spot[V];
	if(uv[U]*temp[U]+uv[V]*temp[V]>0) {
		uv[U]=-uv[U];
		uv[V]=-uv[V];
	}
}

void FinishUVPtByDelta(PMODEL m,uvpt* u,PDELT d) {
	u->Index=d->Index;
	u->Spot[U]=d->Data[0].f;
	u->Spot[V]=d->Data[1].f;
}

void PullUVData(face* f,edge* e) {
	face* of;
	uvpt* u;
	vert *v;

	if(f==NULL) return; // invalid call
	if(e==NULL) return; // invalid call

	if(e->Flags&TF_SHARP) return; // no pull over sharp edges

	of=OtherSide(e,f);
	if(of==NULL) return; // no pull over border edges

	if(of->MapA==NULL) return; // no pull from non-uvmapped faces

	if(f->Texture!=of->Texture) return; // no pull from face with a different texture

	u=NULL;
	v=e->VertS;
	if(v==of->VertA) u=of->MapA;
	if(v==of->VertB) u=of->MapB;
	if(v==of->VertC) u=of->MapC;
	if(v==of->VertD) u=of->MapD;

	if(v==f->VertA) f->MapA=u;
	if(v==f->VertB) f->MapB=u;
	if(v==f->VertC) f->MapC=u;
	if(v==f->VertD) f->MapD=u;

	u=NULL;
	v=e->VertE;
	if(v==of->VertA) u=of->MapA;
	if(v==of->VertB) u=of->MapB;
	if(v==of->VertC) u=of->MapC;
	if(v==of->VertD) u=of->MapD;

	if(v==f->VertA) f->MapA=u;
	if(v==f->VertB) f->MapB=u;
	if(v==f->VertC) f->MapC=u;
	if(v==f->VertD) f->MapD=u;
}

void RemMapping(face* f,BOOL verify) {
	if(f==NULL) return;

	f->MapA = f->MapB = f->MapC = f->MapD = NULL;
	RemoveUnusedUVPts(NULL);
}

void RemoveUnusedUVPts(PMODEL model) {
	MYPTR m,nm;

	if(model==NULL) model=(PMODEL)ActiveProject;

	STEPUVPTS(model,m.u)
	m.u->Flags&=~TF_NEW;

	STEPFACES(model,m.f)
	if(m.f->MapA) {
		m.f->MapA->Flags|=TF_NEW;
		m.f->MapB->Flags|=TF_NEW;
		m.f->MapC->Flags|=TF_NEW;
		if(m.f->MapD) m.f->MapD->Flags|=TF_NEW;
	}

	m.u=FIRSTUVPT(model);
	while(NEXT(m.u)) {
		nm.u=(uvpt*)NEXT(m.u);
		if((m.u->Flags&TF_NEW)==0)
			DestroyUVPt(m.u,FALSE);
		m.u=nm.u;
	}
}

void snapUVPt(uvpt* u,float f) {
	u->Spot[U]=floor((double)u->Spot[U]/f+.5)*f;
	u->Spot[V]=floor((double)u->Spot[V]/f+.5)*f;
}

