#include <math.h>
#include "data.h"

extern PPROJECT ActiveProject;
extern short HighType;
extern MYPTR HighItem;
extern HWND hwndMain;

void UnleashVertex(vert* v) {
	if(v!=NULL) v->Owner=0;
}

vert* CreateVertex(PMODEL m,short pX,short pY,BOOL make_delta) {
	vert* v;

	if(m==NULL) m=(PMODEL)ActiveProject;

	v=(vert*)malloc(sizeof(vert));
	if(v) {
		memset(v,0,sizeof(vert));
		addTail(&m->Vertices,(link*)v);
//		v->Select=0;
//		v->Flags=0;
//		v->Owner=NULL;
//		v->Copy=NULL;
		mapScr2XYZ(&((PPROJECT)m)->cView,pX,pY,v->Spot);
		v->Index=m->NextVertIndex++;
		if(make_delta) CreateVertDelta(v,DT_ADD);
	}
	return v;
}

PDELT CreateVertDelta(vert* v,TYPE dtype) {
	PMODEL m;
	PDELT d;
	LSMWORD ext[2];

	m=FindModel((link*)v);

	d=CreateDelta((PPROJECT)m,dtype,VERTDELT,0);
	if(d==NULL) {
		ReportSystemError("Creation of Delta object failed.");
		return d;
	}
	d->Index=v->Index;

	if(dtype==DT_DEL) return d;

	ExtendDelta(d,3,(LSMWORD *)&v->Spot);

	if(v->Owner) {
		ext[0].l=v->Owner->Index;
	} else {
		ext[0].l=(INDEX)-1;
	}
	ExtendDelta(d,1,ext);

	if(v->Flags&VF_WITHB) {
		ext[0].l=ATBASEFLAG;
		ExtendDelta(d,1,ext);
	}
	if(v->Flags&VF_LOCKED) {
		ext[0].l=LOCKVFLAG;
		ExtendDelta(d,1,ext);
	}

	return d;
}

void DestroyVertex(vert* v,BOOL make_delta) {
	edge *e,*ne;
	face *f,*nf;
	PMODEL m;

	if(v==NULL) {
 		ReportSystemError("NULL pointer passed to DestroyVertex()");
 		return;
	}

	m=FindModel((link*)v);

	f=FIRSTFACE(m);
	while(f!=NULL) {
		nf=(face*)NEXT(f);
		if(f->VertA==v || f->VertB==v || f->VertC==v || f->VertD==v)
		DestroyFace(f,FALSE);//make_delta);
		f=nf;
	}

	e=FIRSTEDGE(m);
	while(e!=NULL) {
		ne=(edge*)NEXT(e);
		if(e->VertS==v || e->VertE==v) DestroyEdge(e,FALSE);//make_delta);
		e=ne;
	}

//	if(make_delta) CreateVertDelta(v,DT_DEL);

	if(HighItem.v==v) {
		HighItem.v=NULL;
		HighType=HT_NOTHING;
	}

	((PPROJECT)m)->Flags|=MF_MESHCHANGED;
	remLink((link*)v);
	free(v);
//	AddAtomDelta(ActiveProject);
}

void DetachVertex(vert* v,BOOL make_delta) {

	v->Owner=NULL;
	v->Flags&=~VF_WITHB;
	if(make_delta) {
//		CreateVertDelta(v,DT_CHG);
//		AddAtomDelta(ActiveProject);
	}
}

void FinishVertByDelta(PMODEL m,vert* v,PDELT d) {
	UINT i;

	v->Index=d->Index;
	v->Spot[X]=d->Data[0].f;
	v->Spot[Y]=d->Data[1].f;
	v->Spot[Z]=d->Data[2].f;
	v->Owner=IndexPart(m,d->Data[3].l);
	// get attachment data
	i=4;
	while(i<d->Len)
	switch(d->Data[i].l) {
		case ATBASEFLAG:
			v->Flags|=VF_WITHB;
			i++;
		break;
		case LOCKVFLAG:
			v->Flags|=VF_LOCKED;
			i++;
		break;
		default: i++;
	}
}

void SnapVertex(vert* v) {
	vec tmp;

	if(v->Owner) {
		if(v->Flags&VF_WITHB) {
			InvertVector(v->Owner->mB,v->Spot,tmp);
		} else {
			InvertVector(v->Owner->mE,v->Spot,tmp);
		}
	} else {
		tmp[X]=v->Spot[X];
		tmp[Y]=v->Spot[Y];
		tmp[Z]=v->Spot[Z];
	}

	tmp[X]=floor((double)tmp[X]/((PMODEL)ActiveProject)->MeshSnap+.5)
		*((PMODEL)ActiveProject)->MeshSnap;
	tmp[Y]=floor((double)tmp[Y]/((PMODEL)ActiveProject)->MeshSnap+.5)
		*((PMODEL)ActiveProject)->MeshSnap;
	tmp[Z]=floor((double)tmp[Z]/((PMODEL)ActiveProject)->MeshSnap+.5)
		*((PMODEL)ActiveProject)->MeshSnap;

	if(v->Owner) {
		if(v->Flags&VF_WITHB) {
			TransformVector(v->Owner->mB,tmp,v->Spot);
		} else {
			TransformVector(v->Owner->mE,tmp,v->Spot);
		}
	} else {
		v->Spot[X]=tmp[X];
		v->Spot[Y]=tmp[Y];
		v->Spot[Z]=tmp[Z];
	}
}

/*
void ReplaceVertex(vert* oldv,vert* newv) {
	edge* e;
	face* f;

	STEPEDGES(ActiveProject,e) {
		if(e->VertS==oldv) { e->VertS=newv; RefreshEdge(e); }
		if(e->VertE==oldv) { e->VertE=newv; RefreshEdge(e); }
	}

	STEPFACES(ActiveProject,f) {
		if(f->VertA==oldv) f->VertA=newv;
		if(f->VertB==oldv) f->VertB=newv;
		if(f->VertC==oldv) f->VertC=newv;
		if(f->VertD==oldv) f->VertD=newv;
	}
}
*/
